#
'''
model.register_forward_pre_hook 会在模型forward之前调用
model.register_forward_hook 会在模型forward之后调用

-------------
之前写的不准确的：
model.register_full_backward_hook 会在模型反向梯度计算之后调用，注，模型列表第一个模型会在计算梯度前调用，该hook参数是(model, grad_ins, grad_outs),model是当前模型，grad_outs是后面的模型传来的，grad_ins是grad_outs经过该model计算后往前传的梯度，好像是第一个模型因为不用往前传梯度，提前调用了

修改后：
model.register_full_backward_hook 会在模型反向梯度计算之后调用，注，当前模型如果不需要再往前传送梯度，会在反向梯度计算前调用该方法，该hook参数是(model, grad_ins, grad_outs),model是当前模型，grad_outs是后面的模型传来的，grad_ins是grad_outs经过该model计算后往前传的梯度

导致的后果是调用该方法的时候，计算数据还没存到显卡，需要延后调用
-------------


torch.autograd.graph.saved_tensors_hooks(hook_pack, hook_unpack):
    hook_pack会在Tensor计算完成后把之后反向梯度计算要用的tensor进行存储
    hook_unpack是在Tensor反向梯度计算前取回之前存储的tensor，注：调用loss.backward不用在with ...saved_tensors_hooks(..)中
'''

import torch
from ... import Base
from ... import pyz
class MiddleBase(Base):
    def init(self, nets):
        self.hook(nets)
    def before_forward(self):
        pass
    def after_forward(self):
        pass
    def before_backward(self):
        pass
    def after_backward(self):
        pass
    def hook_forward_before(self, model, ins):
        return ins.cpu()
    def hook_forward_after(self, model):
        pass
    def hook_backward_after(self, model):
        pass
    def tensor_save(self, tuple_data):
        obj = tuple_data
        return obj
    def tensor_load(self, obj):
        tuple_data = obj
        return tuple_data
    # 下面是不应修改的部分
    def hook(self, nets):
        net = nets[0]
        # self.start_id = id(net)
        # self.start_net = net
        hooks = []
        hooks += [net.register_forward_pre_hook(self.hook_forward_before) for net in nets]
        hooks+=[net.register_forward_hook(self.native_hook_forward_after) for net in nets]
        if hasattr(nets[0], "register_full_backward_hook"):
            hooks+=[net.register_full_backward_hook(self.native_hook_backward_after) for net in nets]
        else:
            hooks+=[net.register_backward_hook(self.native_hook_backward_after) for net in nets]
        self.hooks = hooks
        self.wait_backup_models = []
    def unhook(self):
        [hook.remove() for hook in self.hooks]
    def do_forward(self, fc):
        with self.wrap_forward():
            return fc()
    def do_backward(self, fc):
        with self.wrap_backward():
            return fc()
    def backupable(self, model):
        return True
    def wrap_backward(self):
        def wrap_enter():
            self.wait_backup_models = []
            self.before_backward()
        def wrap_out(exc_type, exc_val, exc_tb):
            #self.hook_backward_after(self.start_net)
            #print(f"BACK OUT MODELS:", self.wait_backup_models)
            for model in self.wait_backup_models:
                self.hook_backward_after(model)
            self.after_backward()
        return pyz.With(wrap_enter, wrap_out, True)
    def wrap_forward(self):
        obj = torch.autograd.graph.saved_tensors_hooks(self.tensor_save, self.tensor_load)
        def wrap_enter():
            self.before_forward()
            obj.__enter__()
        def wrap_out(exc_type, exc_val, exc_tb):
            obj.__exit__(exc_type, exc_val, exc_tb)
            self.after_forward()
        return pyz.With(wrap_enter, wrap_out, True)
    def native_hook_backward_after(self, model, grad_up, grad_src):
        # if id(model)==self.start_id:
        #     return
        if not self.backupable(model):
            self.wait_backup_models.append(model)
            return
        self.hook_backward_after(model)
    def native_hook_forward_after(self, model, ins, outs):
        self.hook_forward_after(model)
pass



from buildz.gpuz.torch.middle_cache import MiddleCache
import torch
from torch import nn,optim
from torch.utils.data import DataLoader, Dataset
class TestDataset(Dataset):
    def __init__(self, num, dims):
        self.num = num
        self.dims = dims
        self.datas = torch.rand(num, dims)
        self.targets = torch.rand(num, dims)
    def __len__(self):
        return self.num
    def __getitem__(self, i):
        return self.datas[i], self.targets[i]
dims = 64
"""
cd D:\rootz\python\gits\buildz_upd\buildz\gpuz\test
D:
python -m buildz.gpuz.test.demo > demo.txt

"""
dataset = TestDataset(90, dims)
dataloader = DataLoader(dataset, 30)
class MiniModel(nn.Module):
    def __init__(self, dims, mdims, num):
        super().__init__()
        nets = [nn.Linear(dims, mdims)]
        nets += [nn.Linear(mdims,mdims) for i in range(num)]
        nets.append(nn.Linear(mdims,dims))
        self.nets = nn.Sequential(*nets)
    def forward(self, inputs):
        return self.nets(inputs)
models = [MiniModel(dims, 32, 3) for i in range(6)]
#opts = [optim.Adam(model.parameters(), lr=0.001) for model in models]
real_model = nn.Sequential(*models)
opt = optim.Adam(real_model.parameters(), lr=0.001)
#可以指定哪些模型全部放cuda或者全部放cpu
cuda_models = [models[1],models[2]]
cpu_models = [models[-1]]
loss_fn = torch.nn.MSELoss()
cache = MiddleCache(models, 3)#, cal_dv = "cpu", cal_nets = cuda_models, cache_nets = cpu_models)#, [torch.device('cuda'), torch.device('cpu')],models,opts,3,opt_step)#, [cuda_models,cpu_models])

# 训练:
def train():
    cache.train()
    _loss = 0
    for inputs,targets in dataloader:
        targets = targets.cuda()
        #inputs,targets = inputs.cuda(),targets.cuda()
        opt.zero_grad()
        with cache.wrap_forward():
            outs = real_model(inputs)
        loss = loss_fn(outs, targets)
        with cache.wrap_backward():
            loss.backward()
        torch.nn.utils.clip_grad_norm_(real_model.parameters(), max_norm=1.0)
        opt.step()
        _loss+=loss.cpu().item()
        #cache.do_backward(lambda: loss.backward())
        # opt.step()在do_backward里会自动调用
        # print(loss.item())
        # break
        #break
    return _loss / (len(dataloader))
pass
import time
print("start")
start = time.time()
for i in range(50):
    print(train())
sec = time.time()-start
print("time cost:", sec)

"""
python -m buildz.gpuz.test.dm1

from buildz.gpuz.test import dm1 as dm

dm.cache.train()
inputs, targets = list(dm.dataloader)[0]
dm.opt.zero_grad()
with dm.cache.wrap_forward():
    outs = dm.real_model(inputs)

pass

ks = list(dm.cache.datas.keys())
dts0 = dm.cache.datas[ks[0]]
net0 = dm.cache.nets[ks[0]]
[k.shape for k in dts0]


dts1 = dm.cache.datas[ks[1]]
net1 = dm.cache.nets[ks[1]]
[k.shape for k in dts1]


ws = []
for md in dm.models:
    for net in md.nets:
        ws.append(net.weight.cpu().detach().clone())

pass

dm.train()

nws = []
for md in dm.models:
    for net in md.nets:
        nws.append(net.weight.cpu().detach().clone())

pass
subs = [(a-b).sum() for a,b in zip(ws, nws)]

w0 = dm.models[0].nets[-1].weight.cpu().detach()
w1 = dm.models[1].nets[0].weight.cpu().detach()

dm.train()

nw0 = dm.models[0].nets[-1].weight.cpu().detach()
nw1 = dm.models[1].nets[0].weight.cpu().detach()

(w0-nw0).sum()
(w1-nw1).sum()

"""


#coding=utf-8
# from buildz.tools import *
from . import pyz,xf,fz,Base, ioc, logz,xz, pathz, cachez, argz, evalz,dz,confz, path as pathx,argx, log as logx, confx
from .tz.time import timecost,showcost
from .ioc import wrap
from .base import fcBase



from buildz.gpuz.torch import DictCache
import torch
from torch import nn,optim
from torch.utils.data import DataLoader, Dataset
class TestDataset(Dataset):
    def __init__(self, num, dims):
        self.num = num
        self.dims = dims
        self.datas = torch.rand(num, dims)
        self.targets = torch.rand(num, dims)
    def __len__(self):
        return self.num
    def __getitem__(self, i):
        return self.datas[i], self.targets[i]
dims = 64
"""
cd D:\rootz\python\gits\buildz_upd\buildz\gpuz\test
D:
python -m buildz.gpuz.test.demo > demo.txt

"""
dataset = TestDataset(90, dims)
dataloader = DataLoader(dataset, 30)
class MiniModel(nn.Module):
    def __init__(self, dims, mdims, num):
        super().__init__()
        nets = [nn.Linear(dims, mdims)]
        nets += [nn.Linear(mdims,mdims) for i in range(num)]
        nets.append(nn.Linear(mdims,dims))
        self.nets = nn.Sequential(*nets)
    def forward(self, inputs):
        return self.nets(inputs)
models = [MiniModel(dims, 32, 3) for i in range(6)]
opts = [optim.Adam(model.parameters(), lr=0.001) for model in models]
#可以指定哪些模型全部放cuda或者全部放cpu
cuda_models = [models[1],models[2]]
cpu_models = [models[-1]]
real_model = nn.Sequential(*models)
loss_fn = torch.nn.MSELoss()
def opt_step(net, opt):
    # 如果模型只是用来测试，不做训练，可以不传该函数，同时opts传入空就可以
    # 对模型的一些其他优化，可以写可以不写，主要是调用opt.step()进行当前小模型的模型训练
    # 另外，opt不一定就是优化函数，可以是任何数据，其只取决于创建DictCache的时候传入的opts是什么
    torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
    opt.step()
cache = DictCache([torch.device('cuda'), torch.device('cpu')],models,opts,3,opt_step)#, [cuda_models,cpu_models])

# 训练:
def train():
    [md.train() for md in models]
    for inputs,targets in dataloader:
        inputs,targets = inputs.cuda(),targets.cuda()
        [opt.zero_grad() for opt in opts]
        outs = cache.do_forward(lambda:real_model(inputs))
        loss = loss_fn(outs, targets)
        cache.do_backward(lambda: loss.backward())
        # opt.step()在do_backward里会自动调用
        #print(loss.item())
        #break
pass
import time
print("start")
start = time.time()
for i in range(50):
    train()
sec = time.time()-start
print("time cost:", sec)
"""
from buildz.gpuz.test import dm

ws = []
for md in dm.models:
    for net in md.nets:
        ws.append(net.weight.cpu().detach().clone())

pass

dm.train()

nws = []
for md in dm.models:
    for net in md.nets:
        nws.append(net.weight.cpu().detach().clone())

pass
subs = [(a-b).sum() for a,b in zip(ws, nws)]

"""


#coding=utf-8
import base64
import hashlib
try:
    import cryptography
except ModuleNotFoundError as exp:
    raise Exception("require package cryptography, install it by 'python3 -m pip install cryptography'")
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, padding
import os
def xors(a,b):
    try:
        import numpy as np
        na = np.array(bytearray(a))
        nb = np.array(bytearray(b))
        rst = na^nb
        return rst.tobytes()
    except ModuleNotFoundError:
        rst = [i^j for i,j in zip(a,b)]
        return bytes(rst)

'''
字节流处理
'''
class ByteStream:
    """
        md5 hash + aes 加密
        加密方式大概逻辑：
            datas = (datas || hash(datas)) ^ base[:size]
            base = base[size:]+aes(md5(base))
            return datas
        这里其实应该新建个线程，不断生成密码流让密码流达到固定大小，达到提前生成密码流，节省时间的目的，否则和实时加密区别不大，但先不管了
    """
    def __init__(self, pwd = None, base = None, iv = None, crypt_hash = 32):
        self.cipher = None
        self.crypt_hash = crypt_hash
        if pwd is not None:
            self.init(pwd, base, iv)
    def init(self, pwd, base, iv):
        if iv is None:
            iv = os.urandom(16)
        if base is None:
            base = os.urandom(32)
        obj = hashlib.md5(pwd)
        self.pwd = obj.hexdigest().encode("ascii")
        self.iv = iv
        self.base = hashlib.md5(base).hexdigest().encode("ascii")
        self.cipher = Cipher(algorithms.AES(self.pwd), modes.CBC(self.iv))
        self.encryptor = self.cipher.encryptor()
        self.codes = b""
        #self.padder = padding.PKCS7(algorithms.AES.block_size).padder()
    def next(self):
        data = self.base
        nxt = hashlib.md5(data).hexdigest().encode("ascii")
        self.base = nxt
        out = self.encryptor.update(data)
        return out
    def encrypt(self, dts):
        if self.crypt_hash>0:
            hash = hashlib.md5(dts).hexdigest().encode("ascii")
            dts += hash[:self.crypt_hash]
        return self.update(dts)
    def decrypt(self, dts):
        dts = self.update(dts)
        if self.crypt_hash>0:
            hash = dts[-self.crypt_hash:]
            dts = dts[:-self.crypt_hash]
            vhash = hashlib.md5(dts).hexdigest().encode("ascii")
            assert hash == vhash
        return dts
    def update(self, dts):
        l = len(dts)
        while len(self.codes)<l:
            self.codes+=self.next()
        tmp = self.codes[:l]
        self.codes = self.codes[l:]
        rst = xors(dts, tmp)
        return rst
        # src = np.array(bytearray(dts))
        # cds = np.array(bytearray(self.codes[:l]))
        # self.codes = self.codes[l:]
        # rst = src^cds
        # return rst.tobytes()

pass




#
from .. import evalz,xf,Base

class VarEval(evalz.Eval):
    def init(self, key, stype = 'list'):
        self.key = key
        self.stype = stype
    def call(self, obj):
        args, maps = obj.args, obj.maps
        if self.stype=='list':
            return args[self.key]
        return maps[self.key]
class VarBuild(evalz.Build):
    def init(self, stype="list"):
        super().init()
        self.stype = stype
    def call(self, data):
        return VarEval(data['data'][0], self.stype)
lb = VarBuild('list')
db = VarBuild('dict')

class SizeEval(evalz.Eval):
    def init(self, key):
        self.key = key
    def call(self, obj):
        args, maps = obj.args, obj.maps
        if self.key in 'l,list,args'.split(","):
            return len(args)
        elif self.key in 'd,dict,maps'.split(","):
            return len(maps)
        assert 0
class SizeBuild(evalz.Build):
    def call(self, data):
        return SizeEval(data['data'][0])
class UnitEval(evalz.Eval):
    def init(self, encape):
        self.encape = encape
    def call(self, obj):
        return self.encape()
class UnitBuild(evalz.Build):
    def init(self, unit):
        super().init()
        self.unit = unit
    def call(self, data):
        conf = [data['type']]+data['data']
        encape, _, find = self.unit.get_encape(conf)
        assert find
        return UnitEval(encape)

sb = SizeBuild()
evalBuilder = evalz.EvalBuilder.Make().set(["l","list","args"], lb).set(['d','dict','maps'], db).set("size,len".split(","), sb)




from .base import *
from ... import dz,pyz
class CallEncape(BaseEncape):
    def init(self, unit, call, args, maps):
        super().init()
        self.unit = unit
        self.params = call, args, maps
    def call(self, params=None, **_):
        call, args, maps = self.params
        call = self.obj(call)
        if type(call)==str:
            call,find = self.unit.get(call)
        assert call is not None
        args = self.elist2obj(args)
        maps = self.edict2obj(maps)
        return call(*args, **maps)
class CallDeal(BaseDeal):
    def init(self):
        super().init()
        conf = BaseConf()
        conf.index(1, 'call', need=1)
        conf.index(2, 'args')
        conf.index(3, 'maps')
        conf.key('call', need=1)
        self.update = conf
    def deal(self, conf, unit):
        assert 'call' in conf
        call = conf['call']
        call = self.get_encape(call, unit)
        args = self.get_elist(unit, conf, 'args',[])
        maps = self.get_edict(unit, conf, 'maps', {})
        return CallEncape(unit, call, args, maps)

pass


#
from buildz import Base, xf
from ..ioc.decorator import decorator,IOCDObj
class Loads(IOCDObj):
    def init(self, conf, vtype = "envs", flush=False):
        if type(conf)==str:
            conf = xf.loads(conf)
        self._type = vtype
        self.conf = conf
        self.flush = flush
    def bind(self, wrap):
        super().bind(wrap)
        self.decorator.add_bind(self)
    def call(self):
        conf = self.decorator.obj
        confs = conf.confs
        maps = {}
        for key, val in self.conf.items():
            key = confs.gid(self.decorator.namespace, key)
            maps[key] = val
        if self._type in ('env', 'envs'):
            if self.flush:
                confs.flush_env(maps)
            xf.deep_fill(maps, conf.envs, 1)
        else:
            if self.flush:
                maps = xf.unflush_maps(maps, confs.id)
            confs.push_vars(maps)

pass

class Loadf(Loads):
    def init(self, fp, vtype="env"):
        conf = xf.loadf(fp)
        super().init(conf, vtype)

pass
decorator.regist("load_conf", Loads)
decorator.regist("loads_conf", Loads)
decorator.regist("loadf_conf", Loadf)


#
from ..ioc.base import Base, EncapeData
from .base import FormatData,FormatDeal
from buildz import xf, pyz
import os
dp = os.path.dirname(__file__)
join = os.path.join
class XfileDeal(FormatDeal):
    """
    配置文件载入xfile/xf:
        {
            id:id
            type:xfile
            filepath: item_conf(fp) 
            # or fp: item_conf(fp)
        }
    简写:
        [[xfile, id], fp]
        [xfile, fp]
        [xf, fp]
    例:
        [xfile, test.js]
        [xfile, [val, test.js]]
    """
    def init(self, fp_lists = None, fp_defaults = None):
        self.singles = {}
        self.sources = {}
        super().init("XfileDeal", fp_lists, fp_defaults, 
            join(dp, "conf", "xfile_lists.js"),
            join(dp, "conf", "xfile_defaults.js"))
    def deal(self, edata:EncapeData):
        sid = edata.sid
        data = edata.data
        conf = edata.conf
        data = self.format(data)
        fp = xf.g(data, filepath=None)
        if fp is None:
            fp = xf.g(data, fp = fp)
        fp = self.get_obj(fp, conf)
        rst = xf.loads(xf.fread(fp))
        return rst

pass




class BufferInput:
    def __init__(self, buf):
        self.buf = buf
    def __call__(self, size=1):
        s = self.buf[:size]
        self.buf = self.buf[size:]
        return s

pass
"""
Buffer:
    缓存操作:
        //add：从数据中获取字符串并直接放入缓存，返回空
        get: 从缓存获取字符串
        rget: 从缓存获取字符串，从右往左拿
        size: 缓存大小
        clean: 清空缓存
        full：获取缓存完整字符串
        read_cache: 从数据中获取字符串并直接放入缓存，返回获取的字符串
    操作数据
        read: 从数据中获取字符串
        clean2read：清空缓存和数据中所有读取过的字符串
    pos: 返回缓存索引，数据读取索引
"""
class BufferBase:
    def pos2str(self, pos):
        assert 0
    def init(self):
        self.buffer_base =0
        self.read_base=0
    def pos(self):
        return self.buffer_base, self.buffer_base+self.size()
    def offsets(self):
        return self.buffer_base, self.read_base
    def read_cache(self, size=1):
        assert 0
    def read(self, size=1):
        assert 0
    def clean2read(self, size=1):
        return self.clean(size)
    def clean(self, read_size=0):
        assert 0
    def size(self):
        assert 0
    def full(self):
        assert 0
    def rget(self, size=1):
        assert 0
    def get(self, size=1):
        assert 0
class Buffer(BufferBase):
    def pos2str(self, pos):
        return "read from file"
    def read_cache(self, size = 1):
        s = self.read(size)
        self.buffer+=s
        self.s_read = self.s_read[len(s):]
        return s
    def read(self, size = 1):
        if self.s_read is None:
            self.s_read = self.input(1)
            self.read_base+=1
        l = len(self.s_read)
        if l<size:
            self.s_read += self.input(size-l)
            self.read_base+=size-l
        rst = self.s_read[:size]
        if self.buffer is None:
            self.buffer = self.s_read[:0]
        return rst
    def init(self):
        super().init()
        self.buffer = None
        self.s_read = None
    def __init__(self, input):
        self.input = input
        self.init()
    def add(self, size):
        if self.buffer is None:
            self.buffer = self.s_read[:size]
        else:
            self.buffer += self.s_read[:size]
    def size(self):
        if self.buffer is None:
            return 0
        return len(self.buffer)
    def full(self, size = 0, right = 1):
        return self.buffer
    def clean2read(self, read_size=1):
        if read_size>0:
            self.s_read = self.s_read[read_size:]
        self.buffer = self.buffer[:0]
        self.buffer_base=self.read_base
    def clean(self, read_size = 0):
        if read_size>0:
            self.s_read = self.s_read[read_size:]
        self.buffer = self.buffer[:0]
        self.buffer_base=self.read_base
    def rget(self, size=1):
        return self.buffer[-size:]
    def get(self, size=1):
        return self.buffer[:size]

pass

def decode(s, coding = 'utf-8'):
    coding = coding.lower()
    xcoding = 'utf-8'
    if coding == 'utf-8':
        xcoding = 'gbk'
    try:
        return s.decode(coding)
    except:
        return s.decode(xcoding)

pass
class StrBuffer(BufferBase):
    def pos2str(self, pos):
        s = self.str[pos[0]:pos[1]+1]
        return s
    def read_cache(self, size = 1):
        s = self.read(size)
        l = len(s)
        self.buffer_size+=l
        self.read_base+=l
        return s
    def read(self, size = 1):
        s = self.str[self.read_base:self.read_base+size]
        if type(s)==bytes:
            s = decode(s)
        return s
    def init(self):
        super().init()
        self.buffer_size = 0
    def __init__(self, s):
        self.str = s
        self.init()
    def add(self, size):
        # x = self.str[self.buffer_base+self.buffer_size:self.buffer_base+self.buffer_size+len(arr)]
        # if x!=arr:
        #     print(f"[ERROR] x:({x}), arr:({arr}), bb: {self.buffer_base}, bs:{self.buffer_size}, rb: {self.read_base}")
        #     print("[["+self.str[self.buffer_base:self.read_base+10]+"]]")
        #     raise Exception("")
        self.buffer_size+=size
    def size(self):
        return self.buffer_size
    def full(self):
        return self.str[self.buffer_base:self.buffer_base+self.buffer_size]
    def clean2read(self, read_size=1):
        self.read_base+=read_size
        self.buffer_base = self.read_base
        self.buffer_size = 0
    def clean(self, read_size = 0):
        self.read_base+=read_size
        self.buffer_base = self.read_base
        self.buffer_size = 0
    def rget(self, size=1):
        size = min(size, self.buffer_size)
        return self.str[self.buffer_base+self.buffer_size-size:self.buffer_base+self.buffer_size]
    def get(self, size=1):
        size = min(size, self.buffer_size)
        return self.str[self.buffer_base:self.buffer_base+size]

pass



#

from .base import *
class ObjectItem(WrapItem):
    def call(self, cls):
        if self.add:
            self.conf['source'] = cls.__module__+"."+cls.__name__
        super().call(cls)
        return cls
class ObjectWrap(WrapBase):
    def call(self, **maps):
        return self.obj(**maps)
    def obj(self, **maps):
        maps['type'] = 'obj'
        return ObjectItem(self, maps, True, True)
    def s_args(self, s):
        list = self.loads(s)
        return self.args(*list)
    def args(self, *list):
        conf = {"args": self.loads_list(list)}
        return WrapItem(self, conf)
    def s_maps(self, s):
        maps = self.loads(s)
        return self.maps(**maps)
    def maps(self, **maps):
        conf = {"maps": self.loads_dict(maps)}
        return WrapItem(self, conf)
    def s_sets(self, s):
        sets = self.loads(s)
        return self.sets(**sets)
    def sets(self, **sets):
        conf = {"sets": self.loads_dict(sets)}
        return WrapItem(self, conf)


#coding=utf-8

__version__="0.7.2"

# 小号多
__author__ = "Zzz, emails: 1174534295@qq.com, 1309458652@qq.com"
__doc__ = '''
使用此代码库发现bug或者有需求需要开发都可联系(QQ或邮箱联系，QQ就是邮箱号)
buildz.xf: 简化的json
buildz.xz: 简化对dict和list的读写
buildz.ioc, buildz.iocz: ioc注入
buildz.netz: 网络相关，包括ssl证书生成，正反向代理，端口映射，抓包
buildz.db: SQL脚本处理工具，基于其他SQL处理库做了使用简化
buildz.html: html文本处理
buildz.gpuz: 机器学习用内存做显存的模型缓存，目前只写了基于pytorch的工具，测试卷积和注意力的效果还行
buildz.fz: 文件处理
buildz.auto: 自动化测试
buildz.logz: 简单的日志工具
buildz.pyz: 简化python系统相关调用
buildz.base: 简化python类代码编写
buildz.pathz: 简化文件路径相关的代码编写 
'''

from .argx import fetch as args
from .base import Base, WBase,Args


#
from ..ioc.base import Base, EncapeData
from .base import FormatData,FormatDeal
from buildz import xf, pyz
import os
dp = os.path.dirname(__file__)
join = os.path.join
class MethodCallDeal(FormatDeal):
    """
    对象方法调用:
        {
            id: id
            type: mcall
            source: 对象id
            mcall|method: 调用方法
            vars: [...]
            args: [...]
            maps: {key:...}
            info: item_conf，额外引用信息，默认null
        }
    简写:
        [[mcall, id], source, method, args, maps, info]
        [mcall, method]
    
    例:
        [mcall, obj.test, run] // 调用
    """
    def init(self, fp_lists = None, fp_defaults = None):
        self.singles = {}
        self.sources = {}
        super().init("MethodCallDeal", fp_lists, fp_defaults, 
            join(dp, "conf", "mcall_lists.js"),
            join(dp, "conf", "mcall_defaults.js"))
    def _deal(self, edata:EncapeData):
        sid = edata.sid
        data = edata.data
        conf = edata.conf
        data = self.format(data)
        src = edata.src
        source = xf.g1(data, source=None, src=None)
        method = xf.g1(data, mcall=0, method=0)
        #method = xf.g(data, method=0)
        info = xf.g(data, info=None)
        if info is not None:
            info = self.get_obj(info, src = edata.src, info = edata.info)
        einfo = edata.info
        if type(einfo)==dict and type(info) == dict:
            xf.fill(einfo, info, 1)
        if info is None:
            info = einfo
        if source is not None:
            source = conf.get_obj(source, info = info)
        if source is None:
            source = src
        if source is None:
            raise Exception(f"not object for method {method}")
        if src is None:
            src = source
        method = self.get_obj(method, conf, src)
        if type(method) in [str, bytes]:
            method = getattr(source, method)
        iargs, imaps = None, None
        ivars = None
        if type(info) == dict:
            iargs, imaps, ivars = xf.g(info, m_args = None, m_maps = None, vars=None)
        args = xf.g(data, args=[])
        maps = xf.g(data, maps ={})
        vars = xf.g(data, vars={})
        vars = {k:self.get_obj(v, conf, src) for k,v in vars.items()}
        if ivars is not None:
            xf.fill(ivars, vars, 1)
        ivars = vars
        if iargs is not None:
            args = iargs
        if imaps is not None:
            xf.fill(imaps, maps, 1)
        # args = [self.get_obj(v, conf, src, edata.info) for v in args]
        # maps = {k:self.get_obj(maps[k], conf, src, edata.info) for k in maps}
        self.push_vars(conf, ivars)
        args = [self.get_obj(v, conf, src) for v in args]
        maps = {k:self.get_obj(maps[k], conf, src) for k in maps}
        rst = method(*args, **maps)
        self.pop_vars(conf, ivars)
        return rst

pass




from buildz.threads import *
import time
tasks = SimpleTasks()
manager = RangeManager(1,3)
pool = Pool(tasks, manager)
print(pool)
import threading,time
def test(sec):
    _id = threading.get_ident()
    print(f"thread {_id} [start]")
    time.sleep(sec)
    print(f"thread {_id} stop")

pass
pool.add_task(FcTask(test, [0.5]))
pool.start()
for i in range(10):
    pool.add_task(FcTask(test, [0.5]))


time.sleep(2.0)




from ..base import Base
from . import mapz
class Mapz(Base):
    def init(self, spt = ".", code='utf-8'):
        self.spt = spt
        self.map = {}
        self.code = code
    def keys(self, key):
        if type(key) not in (list, tuple):
            if type(key)==bytes:
                key = key.decode(self.code)
            if type(key)!=str:
                key = str(key)
            key = key.split(self.spt)
        return key
    def get(self, keys):
        keys = self.keys(keys)
        val, find = mapz.dget(self.map, keys, default=None)
        if not find:
            raise KeyError(str(keys))
        return val
    def set(self, keys, val):
        keys = self.keys(keys)
        mapz.dset(self.map, keys, val)
    def has(self, keys):
        keys = self.keys(keys)
        return mapz.dhas(self.map, keys)
    def remove(self, keys):
        keys = self.keys(keys)
        if not self.has(keys):
            raise KeyError(keys)
        mapz.dremove(self.map, keys)
    def __getitem__(self, keys):
        return self.get(keys)
    def __setitem__(self, keys, val):
        self.set(keys, val)
    def __contains__(self, keys):
        return self.has(keys)
    def __delitem__(self, keys):
        return self.remove(keys)


'''
格式化报错
'''
class FormatExp(Exception):
    def __init__(self, err, data, s = ""):
        if len(s)==0:
            errs = "Error: {err}, line: {line}, index: {index}".format(err = err, line = data[0], index = data[1])
        else:
            errs = "Error: {err}, line: {line}, index: {index}, content: [{s}]".format(err = err, line = data[0], index = data[1], s = s)
        super(FormatExp, self).__init__(errs)

pass



#


from . import xf, ioc, fz, dz, path as pathx
from .base import Base
import time, sys, threading, os
'''
    新的buildz.logz
'''
class Check(Base):
    def str(self):
        return f"Check<self.whitelist: {self.whitelist}, self.blacklist: {self.blacklist}, self.default_pass: {self.default_pass}>"
    @staticmethod
    def build_conf(conf):
        whitelist, blacklist, default_pass, passes, rejects,default = dz.g(conf, whitelist=None, blacklist=None, default_pass=False, passes=[], rejects = [], default=False)
        default_pass = default_pass or default
        whitelist = whitelist or passes
        blacklist = blacklist or rejects
        return Check(whitelist, blacklist, default_pass)
    def init(self, whitelist=[], blacklist=[], default_pass=False):
        self.whitelist = set(whitelist)
        self.blacklist = set(blacklist)
        self.default_pass = default_pass
    def add_pass(self, tag):
        self.whitelist.add(tag)
        if tag in self.blacklist:
            self.blacklist.remove(tag)
    def remove_pass(self, tag):
        self.whitelist.remove(tag)
    def add_reject(self, tag):
        self.blacklist.add(tag)
        if tag in self.whitelist:
            self.whitelist.remove(tag)
    def remove_reject(self, tag):
        self.blacklist.remove(tag)
    def set_default(self, val):
        self.default_pass = val
    def call(self, tag):
        if tag in self.whitelist:
            return True
        if tag in self.blacklist:
            return False
        return self.default_pass
class Log(Base):
    '''
        日志的基类或壳，需要传入实际输出方法或者实现do_log方法
        提供方法接收参数生成日志格式字符串，然后调用实际输出方法或do_log
    '''
    def build_tags(self, conf):
        tags,shows = dz.g(conf, tags=None, shows=None)
        if tags is not None:
            tags = Check.build_conf(tags)
        if shows is not None:
            shows = Check.build_conf(shows)
        print(f"tags: {tags}")
        print(f"shows: {shows}")
        self.check_tags = tags or self.check_tags
        self.shows = shows or self.shows
        return self
    def show(self, type, on=True):
        if not on:
            return self.unshow(type)
        self.shows.add_pass(type)
    def unshow(self, type):
        self.shows.add_reject(type)
    def call(self, _tag):
        return self.tag(_tag)
    def tag(self, _tag):
        if _tag is not None and type(_tag) != str:
            try:
                if not hasattr(_tag, "__name__"):
                    _tag = _tag.__class__
                _tag = _tag.__module__+"."+_tag.__name__
            except:
                _tag = str(_tag)
        base = self
        if self.base is not None:
            base = self.base
        log = Log(self.shows, _tag, base)
        return log
    def get_tag(self):
        return self._tag
    def init(self, shows = None, tag= None, base = None, lock = False, check_tags=None):
        self.shows=shows
        self._tag = tag
        self.base = base
        self.check_tags = check_tags
        if lock:
            self.lock = threading.Lock()
        else:
            self.lock = None
    def log(self, level, tag, *args):
        if self.shows is not None and not self.shows(level):
            return
        if self.check_tags is not None and not self.check_tags(tag):
            return
        if self.lock is not None:
            with self.lock:
                return self.do_log(level, tag, *args)
        return self.do_log(level, tag, *args)
    def xlog(self, level, *args):
        if not self.show(level):
            return
        self.log(level, self.tag, *args)
    def do_log(self, level, tag, *args):
        if self.base is not None:
            return self.base.log(level, tag, *args)
        raise Exception("unimpl")
    def __getattr__(self, key):
        def fc(*args):
            self.log(key, self._tag, *args)
        return fc
    def clean(self):
        pass
    def done(self):
        pass

pass
def replaces(s, *args):
    for i in range(0,len(args),2):
        k,v = args[i],args[i+1]
        s = s.replace(k,v)
    return s

pass
def mstr(s):
    if s is None or len(s)==0:
        return s
    return s[:1].upper()+s[1:].lower()
class FormatLog(Log):
    def init(self, shows =None, tag=None, format=None, base=None, lock = False, check_tags=None):
        if format is None:
            format = "[{LEVEL}] %Y-%m-%d %H:%M:%S [{tag}] {msg}\n"
        self.format=format
        super().init(shows, tag, base, lock, check_tags)
    def output(self, msg):
        raise Exception("impl")
    def do_log(self, level, tag, *args):
        m_level = level.lower()
        u_level = level.upper()
        x_level = mstr(level)
        args = [str(k) for k in args]
        msg = " ".join(args)
        if tag is None:
            tag = "base"
        rst = time.strftime(self.format, time.localtime(time.time()))
        msg = replaces(rst, "{Level}", x_level, "{level}", m_level, "{LEVEL}", u_level, "{tag}", tag, "{msg}", msg)
        self.output(msg)
class FpLog(FormatLog):
    @staticmethod
    def build(conf,dp=None):
        fp = dz.g(conf, fp=None)
        dp = dz.g(conf, dp=dp)
        if dp is not None:
            if type(dp)==str:
                dp = pathx.Path(dp)
            fp = dp(fp)
        reset = dz.g(conf, reset=False)
        log = FpLog(fp)
        if reset:
            log.clean()
        log.build_tags(conf)
        return log
    def init(self, fp = None,shows =None, tag=None, format=None, base=None, lock = False, check_tags=None):
        super().init(shows, tag, format, base, lock, check_tags)
        self.fp = fp
    def clean(self):
        fp = time.strftime(self.fp)
        fz.removes(fp)
    def output(self, msg):
        if self.fp is not None:
            fp = time.strftime(self.fp)
            fz.makefdir(fp)
            fz.write(msg.encode("utf-8"), fp, 'ab')

pass
class StdLog(FormatLog):
    @staticmethod
    def build(conf,*a,**b):
        log = StdLog()
        log.build_tags(conf)
        return log
    def output(self, msg):
        sys.stdout.write(msg)

pass
class Logs(Log):
    '''
    {
        type: logs
    }
    '''
    @staticmethod
    def build(conf, *a,**b):
        conf_logs = dz.g(conf, logs=[])
        if len(conf_logs) == 0:
            return StdLog.build(conf)
        logs = []
        for conf_log in conf_logs:
            logs.append(builds(conf_log))
        log = Logs(logs)
        log.build_tag(conf)
        return log
    def init(self, logs=[], shows = None, tag= None, lock = False, check_tags=None):
        super().init(shows, tag, lock=lock, check_tags=check_tags)
        self.logs = list(logs)
    def do_log(self, level, tag, *args):
        for _log in self.logs:
            _log.log(level, tag, *args)
    def done(self):
        for log in self.logs:
            log.done()
    def clean(self):
        for log in self.logs:
            log.clean()

pass
class Builds(Base):
    def init(self, default=None):
        self.fcs = {}
        self.default = default
    def addx(self, **maps):
        for k,v in maps.items():
            self.fcs[k] = v
    def add(self, key, fc):
        self.fcs[key] = fc
    def call(self, conf, type=None,*a,**b):
        if type is None:
            type = dz.g(conf, type=self.default)
        return self.fcs[type](conf,*a,**b)

pass
builds = Builds("logs")
builds.addx(file=FpLog.build)
builds.addx(std=StdLog.build)
builds.addx(logs=Logs.build)
build_conf = builds

def simple(fp=None,std=True, shows=None, tag=None, format=None,lock=False):
    logs = []
    if fp is not None:
        logs.append(FpLog(fp,shows,tag,format,lock))
    if std:
        logs.append(StdLog(shows,tag,format,lock))
    return Logs(logs)

pass
def build(obj=None, shows=None, tag=None, format=None):
    if obj is None:
        return StdLog(shows, tag, format)
    return obj

pass
make=build



from .base import *
from ..ioc_deal import val
class ValDeal(val.ValDeal):
    def init(self):
        super().init()
        conf = BaseConf()
        conf.index(1, 'val', need=1)
        conf.key('val', 'value,data'.split(","), need=1)
        self.update = conf

pass


from buildz import Base, xf
from . import build
class ArgExp(Exception):
    def __repr__(self):
        return self.__str__()
    def __str__(self):
        return f"stype: {self.stype}, trs: {self.trs}, des: {self.des}"
    def __init__(self, stype, trs, des=None, prev = None):
        super().__init__()
        self.prev = prev
        self.stype = stype
        self.trs = trs
        self.des = des

pass
class Args(Base):
    def call(self, args, maps):
        return args, maps
    def deal_exp(self, exp):
        return exp

pass
class ArrArgs(Args):
    def init(self, args = None):
        if args is None:
            args = []
        self.args = args
    def add(self, item):
        self.args.append(item)
        return self
    def call(self, args, maps):
        i=0
        try:
            for i in range(len(self.args)):
                item = self.args[i]
                args, maps = item(args,maps)
        except ArgExp as exp:
            for j in range(i-1, -1, -1):
                exp = self.args[j].deal_exp(exp)
            raise exp
        return args, maps
    def deal_exp(self, exp):
        for i in range(len(self.args)-1, -1, -1):
            exp = self.args[i].deal_exp(exp)
        return exp
class RangeListArgs(Args):
    def init(self, min = 0, max = None):
        self.min = min
        self.max = max
    def deal_exp(self, exp):
        trs = set()
        for key, vtype in exp.trs:
            if vtype=='list':
                trs.add((key+self.min, vtype))
            else:
                trs.add((key, vtype))
        #exp.trs = trs
        return ArgExp(exp.stype, trs, exp.des, exp)
    def call(self, args, maps):
        if len(args)<self.min:
            raise ArgExp("need", set([(self.min, 'list')]))
        if self.max is None:
            args = args[self.min:]
        else:
            args = args[self.min:self.max]
        return args, maps
class RangeListBuild(build.Build):
    def call(self, conf):
        rg = xf.g(conf, range=None)
        if rg is None:
            return None
        if type(rg)==int:
            rg = [rg, None]
        return RangeListArgs(*rg)
class ArgItem(Base):
    # vtype = 'dict' | 'list'
    def init(self, key, vtype="dict", need = False, default = False, value = None, des = None,remove=True):
        self.vtype = vtype
        self.key = key
        self.need = need
        self.default = default
        self.value = value
        self.trs = []
        self.remove = remove
        if des is None:
            des = key
        self.des = des
    def add(self, key, vtype="dict"):
        self.trs.append((key, vtype))
        return self
    def deal_exp(self, exp):
        trs = set()
        for key, vtype in exp.trs:
            if key == self.key and vtype==self.vtype:
                trs.update(self.trs)
            else:
                trs.add((key, vtype))
        #exp.trs = trs
        return ArgExp(exp.stype, trs, exp.des, exp)
        return exp
    def call(self, set_args, args, maps, rst_args, rst_maps):
        val = None
        find = False
        for key, vtype in self.trs:
            if vtype =="dict":
                if xf.dhas(maps, key):
                    val = xf.dget(maps, key)
                    if self.remove:
                        xf.dremove(maps, key)
                    find=True
                    break
                #if key in maps:
                #    val = maps[key]
                #    if self.remove:
                #        del maps[key]
                #    find = True
                #    break
            else:
                if key in set_args:
                    val = args[key]
                    if self.remove:
                        set_args.remove(key)
                    find = True
                    break
        if not find:
            if self.default:
                find = True
                val = self.value
        if not find and self.need:
            raise ArgExp("need", set(self.trs), self.des)
        if self.vtype == 'dict':
            xf.dset(rst_maps, self.key, val)
            #rst_maps[self.key] = val
        else:
            rst_args[self.key] = val

pass
class ArgItemBuild(build.Build):
    def call(self, key, vtype, conf):
        need = xf.g(conf, need=False)
        default = xf.g(conf, default=False)
        value = xf.g(conf, value=None)
        des = xf.g(conf, des=None)
        remove = xf.g(conf, remove=True)
        src = xf.g(conf, src=None)
        srcs = xf.g(conf, srcs = [])
        if src is not None:
            srcs.append(src)
        src = srcs
        item = ArgItem(key, vtype, need, default, value, des, remove)
        for s in src:
            if type(s) not in (list, tuple):
                if type(s)==int:
                    s = [s, 'list']
                else:
                    s = [s, 'dict']
            if s[1] in ('l', 'args'):
                s[1] = 'list'
            elif s[1] in ('d', 'maps'):
                s[1] = 'dict'
            item.add(*s)
        return item
class TrsArgs(Args):
    def deal_exp(self, exp):
        for item in self.args:
            exp = item.deal_exp(exp)
        return exp
    def init(self, args=None, keep = True):
        super().init()
        if args is None:
            args = []
        self.args = args
        self.keep = keep
    def add(self, item):
        self.args.append(item)
        return self
    def call(self, args, maps):
        rst_args, rst_maps = {}, {}
        set_args = set(range(len(args)))
        for item in self.args:
            item(set_args, args, maps, rst_args, rst_maps)
        l_args = 0
        if len(rst_args)>0:
            l_args= max(rst_args.keys())
        out_args = []
        if self.keep and len(set_args)>0:
            for i in range(max(set_args)):
                if i in set_args:
                    out_args.append(args[i])
        for i in range(l_args+1):
            if i in rst_args:
                if len(out_args)<i:
                    raise ArgExp("need", set([(i, 'list')]))
                out_args.append(rst_args[i])
        if self.keep:
            for k in maps:
                if k not in rst_maps:
                    rst_maps[k] = maps[k]
        return out_args, rst_maps

pass
class TrsArgsBuild(build.Build):
    def init(self):
        self.item = ArgItemBuild()
        super().init(self.item)
    def call(self, conf):
        keep = xf.g(conf, keep = False)
        lst = xf.g1(conf, list={}, args={})
        dct = xf.g1(conf, dict={}, maps={})
        items = []
        for k, val in lst.items():
            items.append(self.item(k, 'list', val))
        for k, val in dct.items():
            items.append(self.item(k, 'dict', val))
        if len(items)==0:
            return None
        return TrsArgs(items, keep)
class ArrArgsBuild(build.Build):
    def init(self):
        self.range = RangeListBuild()
        self.trs = TrsArgsBuild()
        super().init(self.trs, self.range)
    def call(self, conf):
        rg = self.range(conf)
        trs = self.trs(conf)
        if rg is None and trs is None:
            return None
        arr = [rg, trs]
        arr = [k for k in arr if k is not None]
        if len(arr)==1:
            return arr[0]
        return ArrArgs(arr)



#
'''
quest
verify

skt.wrap()

skt.input.add()
skt.ouput.add()

Pub(密码1，密码2，数据)

密码(数据)


'''
class Deal(Base):
    def init(self):
        self.deals = {}
    def call(self, skt, server):
        order = skt.get()
        self.deals[order](skt, server)

pass


#
from . import fio
"""
col=1,as_map = 1
{
    a: [a1,a2,a3,...]
    b: [b1,b2,b3,...]
    ...
}

col=1, as_map=0
[
    [a1,a2,a3,...]
    [b1,b2,b3,...]
    ...
]

col=0,as_map = 1
[
    {a:a1,b:b1,...},
    {a:a2,b:b2,...},
    ...
]

col=0,as_map=0
[
    [a1,b1,...],
    [a2,b2,...],
    ...
]
"""
def to_csv(keys, datas, fp, col = True, as_map = False,coding="utf-8"):
    fio.makefdir(fp)
    rst = []
    if col:
        if as_map:
            datas = [datas[k] for k in keys]
        datas = list(zip(*datas))
    else:
        if as_map:
            datas = [[dt[k] for k in keys] for dt in datas]
    datas = [", ".join([str(v) for v in dt]) for dt in datas]
    rst = [", ".join(keys)]+datas
    rs = "\n".join(rst)
    fio.write(rs.encode(coding), fp)

pass

def from_csv(fp, col=True, as_map = False,coding="utf-8"):
    s = fio.read(fp).decode(coding)
    arr = s.split("\n")
    arr = [[v.strip() for v in k.split(",")] for k in arr if k.strip()!=""]
    keys = arr[0]
    datas = arr[1:]
    if col:
        outs = []
        for i in range(len(keys)):
            tmp = [k[i] for k in datas]
            outs.append(tmp)
        if as_map:
            outs = {k:arr for k,arr in zip(keys, datas)}
        datas = outs
    elif as_map:
        datas = [{k:v for k,v in zip(keys, dt)} for dt in datas]
    return keys, datas

pass
def wrap_cols(keys, cols):
    rst = {}
    for key, col in zip(keys, cols):
        rst[key] = col
    return rst

def init_merge_keys(keys_a, keys_b):
    rst = {}
    keys_b = {k:k for k in keys_b}
    for k in keys_a:
        rst[k] = []
    keys_b_ = {}
    for k,v in keys_b.items():
        while v in rst:
            v = v+"_"
        rst[v] = []
        keys_b_[k] = v
    keys_b = keys_b_
    return rst, keys_b
def merge_col_in(datas_a, datas_b, keys, keys_a = None, keys_b = None):
    if type(keys)==str:
        keys = [keys]
    if keys_a is None:
        keys_a = list(datas_a.keys())
    bak_a = keys_a
    keys_a = set(keys_a)
    if keys_b is None:
        keys_b = list(datas_b.keys())
    bak_b = keys_b
    keys_b = set(keys_b)
    l = len(datas_b[keys[0]])
    indexes = {}
    for i in range(l):
        vs = tuple([datas_b[k][i] for k in keys])
        indexes[vs] = i
    rst, keys_b = init_merge_keys(keys_a, keys_b)
    l = len(datas_a[keys[0]])
    for i in range(l):
        vs = tuple([datas_a[k][i] for k in keys])
        if vs not in indexes:
            continue
        j = indexes[vs]
        for k in keys_a:
            rst[k].append(datas_a[k][i])
        for k, k_ in keys_b.items():
            rst[k_].apepnd(datas_b[k][j])
    return rst, [keys_a, keys_b]

def merge_row_in(datas_a, datas_b, keys, keys_a = None, keys_b = None):
    if type(keys)==str:
        keys = [keys]
    if len(datas_a)*len(datas_b)==0:
        return []
    if keys_a is None:
        keys_a = list(datas_a[0].keys())
    bak_a = keys_a
    keys_a = set(keys_a)
    if keys_b is None:
        keys_b = list(datas_b[0].keys())
    bak_b = keys_b
    keys_b = set(keys_b)
    rst, keys_b = init_merge_keys(keys_a, keys_b)
    rst = []
    l = len(datas_b)
    indexes = {}
    for i in range(l):
        dt = datas_b[i]
        vs = tuple([dt[k] for k in keys])
        indexes[vs] = i
    l = len(datas_a)
    for i in range(l):
        dt = datas_a[i]
        vs = tuple([dt[k] for k in keys])
        if vs not in indexes:
            continue
        j = indexes[vs]
        tmp = {}
        for k in keys_a:
            tmp[k] = dt[k]
        dt = datas_b[j]
        for k, k_ in keys_b.items():
            tmp[k_] = dt[k]
        rst.append(tmp)
    keys_b = [v for k,v in keys_b.items()]
    return rst, [keys_a, keys_b]

    
def merge_in(datas_a, datas_b, keys, col = True, keys_a = None, keys_b = None):
    if col:
        return merge_col_in(datas_a, datas_b, keys, keys_a, keys_b)
    return merge_row_in(datas_a, datas_b, keys, keys_a, keys_b)
    



class BaseDeal:
    def has_prev(self):
        return 1
    def has_deal(self):
        return 1
    def labels(self):
        return []
    def types(self):
        return []
    def prepare(self, msg):
        pass
    def regist(self, mgs):
        if self._id is None:
            self._id = mgs.regist()
        self.prepare(mgs)
    def id(self):
        return self._id
    def __str__(self):
        return "BaseDeal"
    def __repr__(self):
        return str(self)
    def __init__(self, *argv, **maps):
        self._id = None
        self.init(*argv, **maps)
    def init(self, *argv, **maps):
        pass
    def deal(self, buffer, rst, mg):
        pass
    def build(self):
        pass

pass




#coding=utf-8
from buildz.tools import *
from buildz.ioc import wrap
@wrap.obj(id="request")
@wrap.obj_args("ref, log", "ref, cache.modify")
class Req(Base):
    def init(self, log, upd):
        self.upd = upd
        self.log = log.tag("Test.Req")
    def call(self, data, fc=None):
        data = self.upd(data)
        self.log.debug(f"test data: {data}")
        if fc is not None:
            return fc(data)
        return True

pass
from buildz.auto import Run
from buildz.ioc.wrap import decorator
import sys
def test():
    #print(f"wrap: {xf.dumps(decorator(),format=1,deep=1)}")
    rst = Run(basedir="res")("data/test")
    print(f"rst: {rst}")

pass
pyz.lc(locals(), test)




'''
http协议模拟
'''
import socket
import select
from buildz import Base, logz
import traceback
# log = logz.simple("./logs/mhttp/%Y%m%d.log")
# 默认日志只打印不存文件
log = logz.simple()
def http_encode(line, headers=None,data_size=0):
    if type(line)==str:
        line = line.encode("utf-8")
    if line[-2:]!=b"\r\n":
        line+=b"\r\n"
    rst = [line]
    if headers is None:
        headers = {}
    if data_size > 0:
        headers['Content-Length'] = data_size
    for key,val in headers.items():
        rst.append(f"{key}: {val}\r\n".encode("utf-8"))
    rst.append(b"\r\n")
    #if data is not None:
    #    rst.append(data)
    rs = b"".join(rst)
    return rs
pass
def http_recv(rfile):
    "out: line(str), headers(dict), raw_data(bytes or null)"
    #rfile = skt.makefile('rb', buffer_size)
    line = rfile.readline()
    if len(line)==0:
        return None,None,None
    #log.debug(f"http_recv: {line}")
    line = line.decode("utf-8")
    headers = {}
    data = None
    dt_size = 0
    while True:
        next = rfile.readline().strip()
        if len(next)==0:
            break
        index = next.find(b": ")
        key = next[:index].decode("utf-8")
        val = next[index+2:].decode("utf-8")
        if key == 'Content-Length':
            val = int(val)
            dt_size = val
        headers[key] = val
    #if dt_size>0:
    #    data = rfile.read(dt_size)
    #rfile.close() # not keep-alive
    return http_decode(line), headers, dt_size
pass
def http_decode(line):
    index = line.find(" ")
    a, line = line[:index], line[index+1:]
    index = line.find(" ")
    b, c = line[:index], line[index+1:]
    #log.debug(f"http_docde: {a,b,c}")
    return a,b,c.strip()
pass
def spt_url(url):
    index = url.find("://")
    prev = url[:index]
    url = url[index+3:]
    addr = url.split("?")[0].split("/")[0]
    url = url[len(addr):]
    if len(url)==0 or url[0]!="/":
        url = "/"+url
    return prev, addr, url

pass
def comb_url(prev, addr, url):
    return prev+"://"+addr+url
def http_encode_send(http_type, url, headers=None, data_size=0, protocol = "HTTP/1.1"):
    prev, addr, url = spt_url(url)
    addrs = addr.split(":")
    ip = addrs[0]
    if prev.upper()=="HTTP":
        port = 80
    elif prev.upper()=="HTTPS":
        port = 443
    else:
        port = 0
    if len(addrs)>1:
        port = int(addrs[1])
    quest= f"{http_type.upper()} {url} {protocol}".encode("utf-8")
    if headers is None:
        headers = {}
    if 'Host' not in headers:
        headers['Host'] = addr
    keep_alive = 'Connection' in headers and headers['Connection'] == 'keep-alive'
    #log.debug(f"http_encode_send: {quest, headers, data_size}")
    bts = http_encode(quest, headers, data_size)
    return bts, (ip, port),keep_alive
def http_encode_rsp(code, rsp_text, headers=None, data_size=0, protocol = "HTTP/1.1"):
    line = f"{protocol} {code} {rsp_text}".encode("utf-8")
    return http_encode(line, headers, data_size)
pass
def chunked_encode(data):
    n = hex(len(data))[2:].encode()
    return n+b"\r\n"+data+b"\r\n"
def readable(skt,timeout=0):
    (rlist,wlist,elist)=select.select([skt],[],[],timeout)
    return len(rlist)>0
def chunked_data(skt):
    dt = skt.readline().strip().decode()
    n = int(dt, 16)
    dt = skt.read(n+2)[:-2]
    return n, dt
def check_chunked(headers):
    chunked = 'Transfer-Encoding' in headers and headers['Transfer-Encoding']=='chunked'
    return chunked
# Base纯粹是代码简写，懒得写__XX__的形式
class WSocket(Base):
    def readable(self):
        return readable(self.skt)
    @staticmethod
    def Connect(addr, buffer_size=-1):
        skt = WSocket(buffer_size)
        skt.connect(addr)
        return skt
    @staticmethod
    def Bind(skt, buffer_size=-1, addr=None):
        wskt = WSocket(buffer_size)
        wskt.bind(skt, addr)
        return wskt
    def init(self, buffer_size=-1):
        self.buffer_size = buffer_size
        w_buffer_size = buffer_size
        if w_buffer_size<0:
            w_buffer_size=0
        self.w_buffer_size = w_buffer_size
    def bind(self, skt, addr=None):
        if addr is None:
            addr = skt.getpeername()
        self.addr = addr
        rfile = skt.makefile('rb', self.buffer_size)
        wfile = skt.makefile('wb', self.w_buffer_size)
        self.skt = skt 
        self.rfile = rfile
        self.wfile = wfile
        self.send = skt.send
        self.recv = skt.recv
        self.write = wfile.write
        self.read = rfile.read
        self.readline = rfile.readline
        self.sendall = skt.sendall
    def connect(self, addr, skt=None):
        if skt is None:
            skt = socket.socket()
            log.debug(f"connect: {addr}")
            skt.connect(tuple(addr))
        self.bind(skt, addr)
    def closefile(self):
        try:
            self.rfile.close()
            self.wfile.close()
        except:
            pass
    def close(self):
        self.closefile()
        try:
            self.skt.close()
        except:
            pass
class SendDone(Base):
    def init(self, skt, to_rel):
        self.skt = skt
        self.to_rel = to_rel
    def call(self):
        if not self.to_rel:
            return
        self.skt.close()
def http_send_head(http_type, url, headers=None, data_size=0,protocol='HTTP/1.1', skt=None, caches=None, fc_connect=None):
    if caches is None:
        caches = {}
    bts, addr,keep_alive = http_encode_send(http_type, url, headers, data_size,protocol=protocol)
    pfx = spt_url(url)[0].lower()
    if fc_connect is None:
        fc_connect = WSocket.Connect
    #log.debug(f"request addr: {addr}")
    #log.debug(f"http_send: {len(bts)}")
    retry=0
    use_caches=skt is None and keep_alive
    need_rel = keep_alive
    addr_key = tuple([pfx]+list(addr))
    if skt is None:
        if not keep_alive or addr_key not in caches:
            skt = fc_connect(addr)
        else:
            skt = caches[addr_key]
            retry=1
    else:
        need_rel=False
    fc_done = SendDone(skt, need_rel)
    try:
        skt.send(bts)
    except:
        if retry:
            skt=fc_connect(addr)
            skt.send(bts)
        else:
            raise
    if use_caches:
        caches[addr_key]=skt
    return skt,fc_done

class HttpMonitor(Base):
    """
        逻辑梳理，模拟http调用，没实际调用
    """
    def close(self):
        for addr, skt in self.skts.items():
            skt.close()
    def init(self,protocol = "HTTP/1.1"):
        self.protocol = protocol
        self.skts = {}
    def send(self, http_type, url, headers=None, data=None,protocol=None, skt=None):
        ldata = 0
        if data is not None:
            if type(data)!=bytes:
                data = str(data).encode("utf-8")
            ldata = len(data)
        else:
            data = b''
        if protocol is None:
            protocol = self.protocol
        skt, fc_done = http_send_head(http_type, url, headers, ldata, protocal,None, self.skts)
        if ldata>0:
            skt.send(data)
        return skt,fc_done
    def chunked_data(self, skt):
        return chunked_data(skt)
    def check_chunked(self, headers):
        return check_chunked(headers)
    def request(self, http_type, url, headers=None, data=None,protocol=None):
        skt,fc_done = self.send(http_type, url, headers, data,protocol)
        line, headers, data_size=http_recv(skt.rfile)
        data = b''
        if data_size>0:
            data = skt.read(data_size)
        if check_chunked(headers):
            while True:
                n, dt = self.chunked_data(skt)
                data+=dt
                if n==0:
                    break
        fc_done()
        return line, headers, data



from .base import *
from ... import dz,pyz
class CVarEncape(BaseEncape):
    def init(self, cvar):
        super().init()
        self.cvar = cvar
    def call(self, params=None, **maps):
        cvar = self.cvar
        if isinstance(cvar, Encape):
            cvar = cvar()
        cvar = self.load(cvar)
        return cvar
class CVarDeal(BaseDeal):
    def init(self):
        super().init()
        conf = BaseConf()
        conf.index(1, 'cvar', need=1)
        conf.key('cvar', 'var,data'.split(","), need=1)
        self.update = conf
    def deal(self, conf, unit):
        assert 'cvar' in conf
        cvar = conf['cvar']
        if Confs.is_conf(cvar):
            cvar = self.get_encape(cvar, unit)
        return CVarEncape(cvar)

pass




class PosCal:
    def __init__(self):
        self.init()
    def init(self):
        self.sizes = [0]
    def cal_offset(self, s):
        n = "\n"
        r = "\r"
        ep = ""
        if type(s)==bytes:
            n = b"\n"
            r = b"\r"
            ep = b""
        s = s.replace(n, ep).replace(r, ep)
        return -len(s)
    def update(self, s):
        #print("update:["+s+"]")
        szs = self.sizes
        n = "\n"
        r = "\r"
        ep = ""
        if type(s)==bytes:
            n = b"\n"
            r = b"\r"
            ep = b""
        arr = s.replace(r, ep).split(n)
        szs[-1]+=len(arr[0])
        arr = arr[1:]
        for v in arr:
            szs.append(len(v))
            #szs[-1]+=len(v)
    def get(self, offset=0):
        #print("cal get:", offset, self.sizes)
        if type(offset) in [bytes, str]:
            offset = -self.cal_offset(offset)
        offset = -offset
        szs = self.sizes
        row = len(szs)-1
        col = szs[row]
        while offset>0 and row>0:
            sz = szs[row]
            if sz >= offset:
                col = sz-offset
                break
            offset -= sz+1
            row -= 1
            col = szs[row]
        #print("cal result:", row, col)
        return [row+1, col+1]


from .. import base
from .. import item
from .. import exp
from . import spt
class SetDeal(spt.PrevSptDeal):
    """
        Map里的key-val读取
    """
    
    def init(self, spt):
        super().init(spt, True,"kv")
    def build(self, obj):
        return obj

pass


#
from ..ioc.base import Base, EncapeData,IOCError
from .base import FormatData,FormatDeal
from buildz import xf, pyz
import os
dp = os.path.dirname(__file__)
join = os.path.join
class BranchDeal(FormatDeal):
    """
        分支branch:
            {
                id: id
                type: branch
                judge: item_conf
                vals: {
                    judge_val: item_conf
                    ...
                }
                default(optional): item_conf
            }
        简写：
            [[branch, id], judge, {vals}, default]
            [branch, judge, {vals}, default]
            [branch, judge, {vals}]
        例:
            [branch, [env, env.test], {true:..., false: ...}]
    """
    def init(self, fp_lists=None, fp_defaults=None):
        super().init("BranchDeal", fp_lists, fp_defaults, join(dp, "conf", "branch_lists.js"), None)
    def deal(self, edata:EncapeData):
        data = edata.data
        data = self.fill(data)
        judge = xf.get_first(data, 'judge', 'key', 'var')
        judge = self.get_obj(judge, edata.conf, edata.src, edata.info)
        vals = xf.get_first(data, 'vals', 'data', 'datas')
        val = None
        if judge in vals:
            val = vals[judge]
            val = self.get_obj(val, edata.conf, edata.src, edata.info)
            return val
        if 'default' not in data:
            raise IOCError(f"<branch> not any vals match '{judeg}'")
        default = data['default']
        val = self.get_obj(default, edata.conf, edata.src, edata.info)
        return val

pass



#
import sys
import os
curr_dir = os.path.dirname(__file__)
if len(sys.argv)==1:
    sys.argv.append("build_ext")
    sys.argv.append('--build-lib')
    sys.argv.append(f"{curr_dir}")
if sys.platform =='win32':
    sys.argv.append("--compiler=mingw32")
from buildz.xf.cpp import setup

"""
mingw32 or linux:
python -m buildz.xf.cpp.setup_mingw32
mscrt or linux
python -m buildz.xf.cpp.setup
"""


#

import sys
from buildz.gpuz.torch import DictCache
import torch,time
from torch import nn,optim
cpu = torch.device('cpu')
cuda = cpu
if torch.cuda.is_available():
    cuda = torch.device('cuda')
def sz(tensor):
    return tensor.element_size()*tensor.nelement()
def unit_sz(net):
    rst = 0
    for key in "weight,bias".split(","):
        if hasattr(net, key):
            rst+=sz(getattr(net,key))
    return rst

pass
dvs = [cpu,cuda]
def show_size(sz):
    u = None
    for unit in "B,KB,MB,GB".split(","):
        u=unit
        if sz<256:
            break
        sz = sz/1024
    return sz, unit

def default_fc_opt(net, opt):
    torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
    opt.step()
def analyze(use_cache, use_cuda, mark_train, loop, fc_gen, dataloader, loss_fn, fc_opt, win_size):
    tmp = list(fc_gen())
    mds = tmp.pop(0)
    gmodel = tmp.pop(0)
    opts = tmp.pop(0)
    gopt = tmp.pop(0)
    dvs_nets = None
    if len(tmp)>0:
        dvs_nets = tmp.pop(0)
    dv = cuda
    if not use_cuda:
        use_cache = False
    cache = None
    if use_cache:
        cache = DictCache([cuda, cpu], mds, opts, win_size, fc_opt, dvs_nets)
    if not use_cuda:
        dv = cpu
    if not use_cache:
        gmodel = gmodel.to(dv)
    s_val = "mean loss"
    if mark_train:
        if not use_cache:
            gmodel.train()
        else:
            [md.train() for md in mds]
    else:
        s_val = "mean value"
        if not use_cache:
            gmodel.eval()
        else:
            [md.eval() for md in mds]
    times = []
    for i in range(loop):
        total_loss = 0
        curr=time.time()
        if mark_train:
            for dt,tgt in dataloader:
                dt=dt.to(dv)
                tgt = tgt.to(dv)
                if not use_cache:
                    gopt.zero_grad()
                    out = gmodel(dt)
                    loss = loss_fn(out, tgt)
                    loss.backward()
                    fc_opt(gmodel, gopt)
                else:
                    [opt.zero_grad() for opt in opts] #写gopt.zero_grad()应该也可以，只是删掉之前计算的梯度
                    out = cache.do_forward(lambda :gmodel(dt)) # 其实只是加了勾子函数，实际的计算还是模型计算
                    loss = loss_fn(out, tgt)
                    cache.do_backward(lambda : loss.backward()) # 加勾子函数
                total_loss+=loss.item()
        else:
            with torch.no_grad():
                for dt,tgt in dataloader:
                    dt=dt.to(dv)
                    if not use_cache:
                        out = gmodel(dt)
                    else:
                        out = cache.do_forward(lambda :gmodel(dt)) # 其实只是加了勾子函数，实际的计算还是模型
                    total_loss+=out.mean().item()
        sec = time.time()-curr
        print(" train:", i, s_val+":", total_loss/len(dataloader), "time:", sec)
        times.append(sec)
    # 清理显存
    del gmodel, gopt, mds, opts
    torch.cuda.empty_cache()
    return times
        
def analyzes(mark_train, loop, fc_gen, dataloader, loss_fn = None, fc_opt = None, win_size=3, modes = ['cuda', 'cache','cpu']):
    '''
        mds, gmodel, opts, gopt = fc_gen()
    '''
    if type(modes)==str:
        modes = modes.split(",")
    if loss_fn is None:
        loss_fn = torch.nn.MSELoss()
    if fc_opt is None:
        fc_opt = default_fc_opt
    if 'gpu' in modes:
        modes = list(modes)+['cuda']
    print(f"modes: {modes}")
    # 正常做法：只用显卡
    if 'cuda' in modes:
        print("No Used DictCache")
        gtimes = analyze(False, True, mark_train, loop, fc_gen, dataloader, loss_fn, fc_opt, win_size)
        gtimes = gtimes[1:] # 第一次计算因为加载cuda一些东西会比较慢，不计算在内
        torch.cuda.empty_cache()
    # 显卡不足的时候：模型只在计算的时候存入显卡显存
    if 'cache' in modes:
        print("Using DictCache:")
        dc_times = analyze(True, True, mark_train, loop, fc_gen, dataloader, loss_fn, fc_opt, win_size)
        dc_times = dc_times[1:]
        torch.cuda.empty_cache()
    # 不用显卡，只用CPU
    if 'cpu' in modes:
        print("Using CPU:")
        cpu_times = analyze(False, False, mark_train, loop, fc_gen, dataloader, loss_fn, fc_opt, win_size)
        cpu_times = cpu_times[1:]
        torch.cuda.empty_cache()
    print(f"\nAnalyze")
    if 'cuda' in modes:
        print(f"    mean time cost not used DictCache: {sum(gtimes)/len(gtimes)} sec")
    if 'cache' in modes:
        print(f"    mean time cost using DictCache: {sum(dc_times)/len(dc_times)} sec")
    if 'cpu' in modes:
        print(f"    mean time cost using CPU: {sum(cpu_times)/len(cpu_times)} sec")


#coding=utf-8
from buildz import xf, argx, pyz, ioc, fz
from os.path import dirname, join
args, maps = argx.fetch()
en = argx.get(maps, 'en', 0)
text_sfx = "" if not en else ".en"
class Help:
    def __init__(self, dp, fp):
        fp = join(dp, fp)
        self.obj = xf.loads(xf.fread(fp))
        self.text = self.obj['text'+text_sfx]
    def deal(self):
        print(self.text)

pass
class Deal:
    def __init__(self, conf, deals, default):
        self.deals = {}
        self.default = default
        for md in deals:
            self.deals[md] = {}
            refs = deals[md]
            for key in refs:
                ref = refs[key]
                obj = conf.get(ref)
                self.deals[md][key] = obj
    def run(self):
        if len(args)==0:
            return self.default.deal()
        md = args[0]
        help = "h" in maps or "help" in maps
        key = "deal" if not help else "help"
        if md not in self.deals:
            return self.default.deal()
        obj = self.deals[md][key]
        obj.deal()

pass
dp = dirname(__file__)
import time
def test():
    print(f"start\n{'='*30}")
    args, maps = argx.fetch()
    curr = time.time()
    fps = fz.search(join(dp, 'res', 'conf'), ".*\.js")
    confs = ioc.build()
    confs.add_fps(fps)
    confs.get("run")
    sec = time.time()-curr
    ps = f"{'='*30}\nPS: 这只是个使用展示demo，代码在buildz.demo里，可供参考，运行耗时: {sec}"
    if en:
        ps = f"{'='*30}\nPS: just an demo show used of buildz module, demo codes is in buildz.demo, run time cost: {sec}"
    print(ps)

pass
    



#

from buildz.argz import conf_argz as argz
from buildz.tls import *
conf = xf.loads(r"""
#range=(0,12,3)
range=(1,12,1)
args: {
    0:[0,0]
}
maps: {
    a={
        default: test
    }
}
""")
def test():
    args = [0,1,2]
    maps = {'a':'b','c':'d'}
    bd = argz.FullArgsBuilder()
    obj = bd(conf)
    args, maps = obj.deal(args, maps)
    print(f"obj: {obj}")
    print(f"args: {args}, maps: {maps}")
    pass

pass

pyz.lc(locals(),test)


from .. import base
from .. import item
from .. import exp
class LRDeal(base.BaseDeal):
    """
        分隔符，有分隔符后将缓存的数据当作字符串
    """
    def init(self, left, right, name= "lr"):
        self.left = left
        self.right = right
        self.ll = len(left)
        self.lr = len(right)
        self.name = name
    def build(self, arr, l_item):
        # implement this
        return item.DealItem(arr, l_item.pos, self.id())
    def err(self, s):
        return s.replace("<lr>", self.name)
    def prev(self, buffer, queue, pos):
        cl = buffer.read(self.ll)
        cr = buffer.read(self.lr)
        if not self.same(self.left, cl) and not self.same(self.right, cr):
            return False
        if self.same(self.left, cl):
            buffer.pop_read(self.ll)
            rm = buffer.full()
            if len(rm.strip())>0:
                pos.update(rm)
                raise exp.FormatExp(self.err("unexcept char before <lr> left symbol"), pos.get(), rm)
            buffer.clean()
            pos.update(cl)
            queue.append(item.PrevItem(cl, pos.get(), self.id(), left = 1))
        else:
            buffer.pop_read(self.lr)
            rm = buffer.full()
            buffer.clean()
            pos.update(rm)
            # 这里对扩展不太好，需要注意
            #if len(rm.strip())>0 or len(queue)==0 or queue[-1].none(is_val = 1,type=self.id(), right=1,allow_empty=0):
            if len(rm.strip())>0:
                queue.append(item.PrevItem(rm, pos.get(), is_val=1))
            pos.update(cr)
            queue.append(item.PrevItem(cr, pos.get(), self.id(), right = 1))
        return True
    def deal(self, queue, stack):
        if len(queue)==0:
            return False
        _item = queue[0]
        if not _item.check(type=self.id()):
            return False
        if not _item.check(left=1) and not _item.check(right=1):
            return False
        queue.pop(0)
        if _item.check(left = 1):
            stack.append(_item)
            return True
        rst = []
        l_item = None
        while len(stack)>0:
            item_i = stack.pop(-1)
            if item_i.check(type = self.id()) and item_i.check(left=1):
                l_item = item_i
                break
            rst.append(item_i)
        if l_item is None:
            raise exp.FormatExp(self.err("find <lr> right symbal while not left symbal found"), _item.pos, _item.val)
        rst.reverse()
        rst = self.build(rst, l_item)
        stack.append(rst)
        return True

pass


#

import re
from .jsc import fetch_methods_text, rps_dct, fetch_vals,fetch_vars

from .util import make_fcs
sc_rename = """
    <type> <method>(<params>)<exp> {
        <ret><call>(<vars>);
    }
""".rstrip()
# def rename(codes, args, text):
#     _type, method, params, has_def = fetch_method_text(codes, text)
#     rst = []
#     if has_def:
#         rst.append(codes.rstrip())
def rename(_type, method, params, exp, args):
    rst = []
    args = fetch_vals(args)
    ret = "return " if _type.find("void")<0 else ""
    wmethod = args[0]
    params = [k.strip() for k in fetch_vars(params)]
    vars = [k.split(" ")[-1].strip() for k in params]
    key = 'key'
    if len(args)>1:
        key =args[1]
    rs = rps_dct(sc_rename, type =_type,method=wmethod,params=', '.join(params), call=method, vars = ', '.join(vars), ret = ret,exp=exp)
    rst.append(rs)
    rst.reverse()
    return rst, []
    
renames = make_fcs(rename, False)



#
from ... import pyz,dz,Base
class TagDict(Base):
    def remove(self, key, tag=None):
        tag = pyz.nnull(tag, self.default)
        rst = dz.get_set(self.maps, tag, dict())
        dz.dremove(rst, key)
    def update(self, maps, tag=None):
        tag = pyz.nnull(tag, self.default)
        rst = dz.get_set(self.maps, tag, dict())
        dz.fill(maps, rst)
    def set(self, key, val, tag=None):
        tag = pyz.nnull(tag, self.default)
        rst = dz.get_set(self.maps, tag, dict())
        dz.dset(rst, key, val)
    def get(self, key, tags=None):
        if tags is None:
            tags = self.maps.keys()
        elif type(tags) not in (list,tuple,set):
            tags = [tags]
        for tag in tags:
            if tag not in self.maps:
                continue
            rst = self.maps[tag]
            val, has = dz.dget(rst, key)
            if has:
                return val, tag, 1
        return None, None, 0
    def tag(self, tag):
        rst = dz.get_set(self.maps, tag, dict())
        return rst
    def __getattr__(self, tag):
        return self.tag(tag)
    def init(self, default=""):
        self.maps = {}
        self.default=default

pass


#
import sys
from buildz.gpuz.test.test_linear_demo import *
def test():
    '''
        占显存用的，直接从test_linear_demo import代码过来完事,调用：
        python -m buildz.gpuz.test.take_gpu_mem
        或
        python -m buildz.gpuz.test.take_gpu_mem num
        num默认12，占用3GB显存
    '''
    nets=10
    dims=2000
    trains = 5
    datas = 60
    batch=30
    lr=0.0001
    win_size=3
    num = 12
    args = sys.argv[1:]
    if len(args)>0:
        num = int(args[0])
    print(f"num: {num}")
    def fc_gen():
        mds = [Model(dims, nets) for i in range(num)]
        mds_sz = [md.size() for md in mds]
        sz, unit = analyze.show_size(sum(mds_sz))
        print(f"Model Size: {sz} {unit}")
        opts =[optim.Adam(md.parameters(), lr=lr) for md in mds]
        gmodel = nn.Sequential(*mds)
        gopt = optim.Adam(gmodel.parameters(), lr=lr)
        return mds, gmodel, opts, gopt
    mds,gmodel,opts,gop=fc_gen()
    ds = TestDataset(datas, dims)
    dl = DataLoader(ds, batch)
    dt = list(dl)[0][0]
    gmodel = gmodel.cuda()
    with torch.no_grad():
        dt=dt.cuda()
        while True:
            out = gmodel(dt)
            print(out.mean())
            time.sleep(1)

pyz.lc(locals(),test)



from .base import *
from ... import dz,pyz,xf
class AttrEncape(BaseEncape):
    def init(self, unit, attr, obj):
        super().init()
        self.unit = unit
        self.params = attr, obj
    def call(self, params=None, **maps):
        attr, obj = self.params
        attr = self.obj(attr)
        if isinstance(obj, Encape):
            obj = self.obj(obj)
        elif type(obj)==str:
            obj, find = self.unit.get(obj)
        if obj is None and params is not None:
            obj = params.obj
        assert obj is not None and hasattr(obj, attr), f"attr get failed for "
        return getattr(obj, attr)
class AttrDeal(BaseDeal):
    def init(self):
        super().init()
        conf = BaseConf()
        conf.ikey(1, 'attr', 'var,data,key'.split(","), need=1)
        conf.ikey(2, 'obj', 'object,src,source'.split(','))
        self.update = conf
    def deal(self, conf, unit):
        assert 'attr' in conf
        attr = conf['attr']
        attr = self.get_encape(attr, unit)
        # obj = dz.g(conf, obj=None)
        # print(f"obj: {obj}")
        # exit()
        obj = dz.get(conf, 'obj')
        obj = self.get_encape(obj, unit)
        return AttrEncape(unit, attr, obj)

pass


#coding=utf-8
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
from cryptography import x509
from cryptography.x509.oid import NameOID
import datetime
from buildz import xf
from buildz import fz
from getpass import getpass
def sign(private_key, bs):
    return private_key.sign(bs, padding.PKCS1v15(), hashes.SHA256())
def verify(public_key, bs, sig):
    public_key.verify(sig,bs, padding.PKCS1v15(), hashes.SHA256())

pass
def encrypt(public_key, bs):
    encrypted = public_key.encrypt(
        bs,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    return encrypted

pass
def decrypt(private_key, bs):
    decrypted = private_key.decrypt(
        bs,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None
        )
    )
    return decrypted

pass
# 生成RSA私钥
def gen_prv(filepath, pwd = None):
    """
        生成pem格式私钥, 存放到filepath文件上
    """
    if pwd == True:
        pwd = getpass("input private_key password:").encode()
    if type(pwd)==str:
        pwd = pwd.encode()
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend()
    )
    # 将私钥序列化为PEM格式
    if pwd is None:
        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )
    else:
        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.BestAvailableEncryption(pwd)
        )
    # 将私钥保存在磁盘上
    with open(filepath, "wb") as f:
        f.write(pem)

pass

def gen_pub(filepath, fp_prv, pwd = None):
    """
        从私钥文件fp_prv生成pem格式公钥, 保存到filepath文件上
    """
    if pwd == True:
        pwd = getpass("input private_key password:").encode()
    if type(pwd)==str:
        pwd = pwd.encode()
    # 从磁盘上读取私钥
    with open(fp_prv, "rb") as f:
        pem_data = f.read()
    # 将PEM格式的私钥反序列化为私钥对象
    private_key = serialization.load_pem_private_key(pem_data, password=pwd, backend=default_backend())
    # 生成RSA公钥
    public_key = private_key.public_key()
    # 将RSA公钥序列化为PEM格式
    pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    # 将RSA公钥保存在磁盘上
    with open(filepath, "wb") as f:
        f.write(pem)

pass

def load_prv(fp_prv, pwd = None):
    # 从磁盘上读取私钥
    if pwd == True:
        pwd = getpass("input private_key password:").encode()
    if type(pwd)==str:
        pwd = pwd.encode()
    pem_data = fz.read(fp_prv, 'rb')
    # 将PEM格式的私钥反序列化为私钥对象
    private_key = serialization.load_pem_private_key(pem_data, password=pwd, backend=default_backend())
    return private_key

pass
def load_pub(fp_pub):
    # 从磁盘上读取公钥
    pem_data = fz.read(fp_pub, 'rb')
    # 将PEM格式的公钥反序列化为公钥对象
    public_key = serialization.load_pem_public_key(pem_data, backend=default_backend())
    return public_key

pass
def get(map, key, default = None):
    if key not in map:
        return default
    return map[key]

pass
def gen_subject(conf = {}):
    # 生成证书的主体名称
    names = [
        # 国家
        x509.NameAttribute(NameOID.COUNTRY_NAME, get(conf, "contry", "CN")),
        # 省份
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME,  get(conf, "provice", "CN")),
        # 城市
        x509.NameAttribute(NameOID.LOCALITY_NAME, get(conf, 'local', 'cityz')),
        # 组织
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, get(conf, 'org', 'orgz')),
        # 这里其实应该是域名/ip
        x509.NameAttribute(NameOID.COMMON_NAME, get(conf, 'comman', 'commanz')),
        # email
        x509.NameAttribute(NameOID.EMAIL_ADDRESS, get(conf, 'email', 'emailz'))
    ]
    subject = x509.Name(names)
    return subject

pass

def add_extensions(builder, conf = {}):
    dns = xf.g(conf, dns = [])
    if len(dns)>0:
        #print("dns:", dns)
        dns_arr = [x509.DNSName(k) for k in dns]
        dns_arr = x509.SubjectAlternativeName(dns_arr)
        builder = builder.add_extension(dns_arr, critical=False)
    ca = xf.g(conf, ca = False)
    if ca:
        builder = builder.add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True
        )
    # DNS
    # builder = builder.add_extension(
    #     x509.SubjectAlternativeName([
    #         x509.DNSName(get(conf, 'dns1', 'localhost')),
    #         x509.DNSName(get(conf, 'dns2', 'localhost'))
    #     ]),
    #     critical=False
    # )
    # builder = builder.add_extension(
    #     x509.BasicConstraints(ca=True, path_length=None),
    #     critical=True
    # )
    return builder

pass
def gen_csr(filepath, fp_prv, conf = {}, pwd = None):
    if pwd == True:
        pwd = getpass("input private_key password:").encode()
    if type(pwd)==str:
        pwd = pwd.encode()
    private_key = load_prv(fp_prv, pwd)
    subject = gen_subject(conf)
    builder = x509.CertificateSigningRequestBuilder()
    builder = builder.subject_name(subject)
    builder = add_extensions(builder, conf)
    # 将证书签名
    certificate = builder.sign(
        private_key=private_key,
        algorithm=hashes.SHA256(),
        backend=default_backend()
    )
    pem = certificate.public_bytes(serialization.Encoding.PEM)
    fz.write(pem, filepath, 'wb')

pass

def load_csr(filepath):
    pem = fz.read(filepath, 'rb')
    csr = x509.load_pem_x509_csr(pem)
    return csr

pass
#要生成证书链，是从网站证书开始写入，往上写到根证书
"""
chains.cert文件结构:
网站证书
中间节点证书
...
根证书
"""
def sign_csr(filepath, fp_csr, fp_cert, fp_prv, conf = {}, pwd = None, add_chains = True):
    # 将PEM格式的私钥反序列化为私钥对象
    if pwd == True:
        pwd = getpass("input private_key password:").encode()
    if type(pwd)==str:
        pwd = pwd.encode()
    private_key = load_prv(fp_prv, pwd)
    cert = load_cert(fp_cert)
    csr = load_csr(fp_csr)
    builder = x509.CertificateBuilder()
    builder = builder.subject_name(csr.subject)
    builder = builder.issuer_name(cert.subject)
    builder = builder.public_key(csr.public_key())
    builder = builder.serial_number(x509.random_serial_number())
    builder = builder.not_valid_before(datetime.datetime.utcnow() - datetime.timedelta(days=xf.g(conf, valid_before=1)))
    builder = builder.not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=xf.g(conf, valid=365)))
    for extension in csr.extensions:
        #print(f"csr excention: {extension}")
        builder = builder.add_extension(extension.value, extension.critical)
    # 将证书签名
    certificate = builder.sign(
        private_key=private_key,
        algorithm=hashes.SHA256(),
        backend=default_backend()
    )
    pem = certificate.public_bytes(serialization.Encoding.PEM)
    if add_chains:
        pem += fz.read(fp_cert, 'rb')
    # 将证书保存在磁盘上
    fz.write(pem, filepath, 'wb')

pass


def gen_cert(filepath, fp_prv, conf={}, pwd = None):
    """
    用私钥fp_prv签名公钥fp_pub和配置conf, 生成证书到filepath
    """
    # 将PEM格式的私钥反序列化为私钥对象
    if pwd == True:
        pwd = getpass("input private_key password:").encode()
    if type(pwd)==str:
        pwd = pwd.encode()
    private_key = load_prv(fp_prv, pwd)
    public_key = private_key.public_key()
    # 将PEM格式的公钥反序列化为公钥对象
    #public_key = load_pub(fp_pub)
    # 生成证书的主体名称
    subject = gen_subject(conf)
    # 生成SSL证书
    builder = x509.CertificateBuilder()
    builder = builder.subject_name(subject)
    issuer = subject
    builder = builder.issuer_name(subject)
    builder = builder.public_key(public_key)
    builder = builder.serial_number(x509.random_serial_number())
    builder = builder.not_valid_before(datetime.datetime.utcnow() - datetime.timedelta(days=xf.g(conf, valid_before=1)))
    builder = builder.not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=xf.g(conf, valid=365)))
    builder = add_extensions(builder, conf)
    # 将证书签名
    certificate = builder.sign(
        private_key=private_key,
        algorithm=hashes.SHA256(),
        backend=default_backend()
    )
    pem = certificate.public_bytes(serialization.Encoding.PEM)
    # 将证书保存在磁盘上
    fz.write(pem, filepath, 'wb')

pass
class CertVerifyException(Exception):
    pass

pass
def loadf_certs(fp):
    bs = fz.read(fp,'rb')
    return load_certs(bs)
def load_certs(bs):
    spt = b"-----END CERTIFICATE-----"
    pems = bs.split(spt)
    pems = [k+spt for k in pems if k.strip()!=b""]
    certs = [x509.load_pem_x509_certificate(pem) for pem in pems]
    return certs
def verify_certs_fp(fp, cas = None, verify_time=True):
    certs = loadf_certs(fp)
    if len(certs)==0:
        raise CertVerifyException(f"not any cert info in {fp}")
    if cas is None:
        cas = []
    if type(cas)!=list:
        cas = [cas]
    cas = [load_cert(ca) for ca in cas]
    return verify_certs(certs, cas, verify_time)

pass
def verify_certs(certs, cas = None, verify_time=True):
    if certs is None:
        certs = []
    if type(certs)!=list:
        certs = [certs]
    if len(certs)==0:
        raise CertVerifyException(f"not any cert info in {fp}")
    if cas is None:
        cas = []
    if type(cas)!=list:
        cas = [cas]
    #cas = [load_cert(ca) for ca in cas]
    roots = {ca.subject:[ca.public_key(), ca.signature_hash_algorithm] for ca in cas}
    #print(f"[TESTZ] verify certs: {len(certs)}")
    for i in range(len(certs)-1):
        curr = certs[i]
        up = certs[i+1]
        verify_cert(curr, up.public_key(), verify_time, up.signature_hash_algorithm)
    cert = certs[-1]
    issuer = cert.issuer
    if issuer not in roots:
        raise CertVerifyException(f"not root cert find: {issuer}")
    pk, hash_alg = roots[issuer]
    verify_cert(cert, pk, verify_time, hash_alg)

pass


def verify_cert(cert, public_key=None, verify_time=True, hash_alg = None):
    if type(cert)==str:
        cert = load_cert(cert)
    if public_key is None:
        public_key = cert.public_key()
    if hash_alg is None:
        hash_alg = cert.signature_hash_algorithm
    public_key.verify(cert.signature,cert.tbs_certificate_bytes, padding.PKCS1v15(),hash_alg)
    if not verify_time:
        return
    if cert.not_valid_before_utc<datetime.datetime.now(datetime.timezone.utc)<cert.not_valid_after_utc:
        return
    raise CertVerifyException("date not pass")

pass

def load_cert(fp):
    bs = fz.read(fp, 'rb')
    cert = x509.load_pem_x509_certificate(bs)
    return cert

pass
def load_der(fp):
    bs = fz.read(fp, 'rb')
    cert = x509.load_der_x509_certificate(bs)
    return cert

pass
SOIDS = "COUNTRY_NAME,STATE_OR_PROVINCE_NAME,LOCALITY_NAME,ORGANIZATION_NAME,COMMON_NAME,EMAIL_ADDRESS".split(",")
OIDS = ['BUSINESS_CATEGORY', 'COMMON_NAME', 'COUNTRY_NAME', 'DN_QUALIFIER', 'DOMAIN_COMPONENT', 'EMAIL_ADDRESS', 'GENERATION_QUALIFIER', 'GIVEN_NAME', 'INITIALS', 'INN', 'JURISDICTION_COUNTRY_NAME', 'JURISDICTION_LOCALITY_NAME', 'JURISDICTION_STATE_OR_PROVINCE_NAME', 'LOCALITY_NAME', 'OGRN', 'ORGANIZATIONAL_UNIT_NAME', 'ORGANIZATION_IDENTIFIER', 'ORGANIZATION_NAME', 'POSTAL_ADDRESS', 'POSTAL_CODE', 'PSEUDONYM', 'SERIAL_NUMBER', 'SNILS', 'STATE_OR_PROVINCE_NAME', 'STREET_ADDRESS', 'SURNAME', 'TITLE', 'UNSTRUCTURED_NAME', 'USER_ID', 'X500_UNIQUE_IDENTIFIER']
#oids = [getattr(NameOID, oid) for oid in OIDS if hasattr(NameOID, oid)]

NAMES = ['业务类别', '公共名称', '国家名称', ' DN_限定符', '域_组件', '电子邮件_地址', '代_限定符', '名字', '旅馆', '管辖区_国家名称', '管辖区_地点_名称', '管辖区_州_或_省_名称', '地点_名称', 'OGRN', '组织_单位_名称', '组织_标识符', '组织_名称', '邮政_地址', '邮政_代码', '假名', '序列号', ' SNILS ', '州_或_省_名称', '街道_地址', '姓氏', '头衔', '非结构化名称', '用户_ID ', 'X500']
names = {getattr(NameOID, OID): f"{NAME}({OID.lower()})" for NAME, OID in zip(NAMES, OIDS) if hasattr(NameOID, OID)}
def des_subject(subject):
    arr = subject._attributes
    rst = {}
    for its in arr:
        for it in its:
            oid = it.oid
            val = it.value
            if oid not in names:
                name = oid.dotted_string
            else:
                name = names[oid]
            rst[name]=val
    return rst

pass
def test():
    gen_prv("prv.pem")
    gen_pub("pub.pem", "prv.pem")
    sign("sign.pem", "pub.pem", "prv.pem")

pass

if __name__=="__main__":
    test()

pass



from .. import base
from .. import item
from .. import exp
from . import lr
class ListDeal(lr.LRDeal):
    """
        分隔符，有分隔符后将缓存的数据当作字符串
    """
    def init(self, left, right):
        super().init(left, right, 'list')
    def types(self):
        return ['list']
    def build(self, obj):
        #print(f"list build: {obj}")
        #if self.check_right(obj):
        #    return None
        val = obj.val
        if len(val)==0:
            obj.val = []
            obj.is_val = 1
            return obj
        if len(val)>1:
            opt = val[1]
            if opt.type=='kv':
                return None
        return self.build_arr(val)
    def build_arr(self, arr):
        rst = []
        for _item in arr:
            if not _item.is_val:
                raise Exception(f"error in list:{_item}")
            rst.append(_item.val)
        return item.Item(rst, type='list', is_val = 1)

pass



from buildz.gpuz.torch import DictCache
import torch
from torch import nn,optim
from torch.utils.data import DataLoader, Dataset
class TestDataset(Dataset):
    def __init__(self, num, dims):
        self.num = num
        self.dims = dims
        self.datas = torch.rand(num, dims)
        self.targets = torch.rand(num, dims)
    def __len__(self):
        return self.num
    def __getitem__(self, i):
        return self.datas[i], self.targets[i]
dims = 3
"""
cd D:\rootz\python\gits\buildz_upd\buildz\gpuz\test
D:
python -m buildz.gpuz.test.demo > demo.txt

"""
dataset = TestDataset(30, dims)
dataloader = DataLoader(dataset, 10)
class MiniModel(nn.Module):
    def __init__(self, dims, mdims, num):
        super().__init__()
        nets = [nn.Linear(dims, mdims)]
        nets += [nn.Linear(mdims,mdims) for i in range(num)]
        nets.append(nn.Linear(mdims,dims))
        self.nets = nn.Sequential(*nets)
    def forward(self, inputs):
        return self.nets(inputs)
models = [MiniModel(dims, 6, 3) for i in range(6)]
opts = [optim.Adam(model.parameters(), lr=0.001) for model in models]
#可以指定哪些模型全部放cuda或者全部放cpu
cuda_models = [models[1],models[2]]
cpu_models = [models[-1]]
real_model = nn.Sequential(*models)
loss_fn = torch.nn.MSELoss()
def opt_step(net, opt):
    # 如果模型只是用来测试，不做训练，可以不传该函数，同时opts传入空就可以
    # 对模型的一些其他优化，可以写可以不写，主要是调用opt.step()进行当前小模型的模型训练
    # 另外，opt不一定就是优化函数，可以是任何数据，其只取决于创建DictCache的时候传入的opts是什么
    torch.nn.utils.clip_grad_norm_(net.parameters(), max_norm=1.0)
    opt.step()
cache = DictCache([torch.device('cuda'), torch.device('cpu')],models,opts,3,opt_step)#, [cuda_models,cpu_models])

# 训练:
[md.train() for md in models]
for inputs,targets in dataloader:
    inputs,targets = inputs.cuda(),targets.cuda()
    [opt.zero_grad() for opt in opts]
    outs = cache.do_forward(lambda:real_model(inputs))
    loss = loss_fn(outs, targets)
    cache.do_backward(lambda: loss.backward())
    # opt.step()在do_backward里会自动调用
    print(loss.item())
    break

exit()

# 测试:
inputs = torch.rand(1, dims).cuda()
with torch.no_grad():
    outputs = cache.do_forward(lambda:real_model(inputs))
print(outputs)

# 对比不用DictCache的时候
# 注意：模型放入DictCache之后会挂上勾子函数，不再用DictCache的时要先调用DictCache的remove方法
cache.remove()
full_opt = optim.Adam(real_model.parameters(), lr=0.001)
# 训练:
real_model.cuda()
real_model.train()
for inputs,targets in dataloader:
    inputs,targets = inputs.cuda(),targets.cuda()
    full_opt.zero_grad()
    outs = real_model(inputs)
    loss = loss_fn(outs, targets)
    loss.backward()
    torch.nn.utils.clip_grad_norm_(real_model.parameters(), max_norm=1.0)
    full_opt.step()
    print(loss.item())

pass

# 测试:
inputs = torch.rand(1, dims).cuda()
with torch.no_grad():
    outputs = real_model(inputs)
print(outputs)




from ..iocz.conf import conf
from .. import Base, xf, dz,pyz
from .argz import *
"""
key:{
    src: [a,b,c]
    des: asdf
    need:1
    default: True
    default_value: 123
    remove: 1
}
key: [
    src, 
    [default, default_value]
    [des, need, remove]
]
"""
class ArgItemConf(conf.Conf):
    def init(self):
        super().init()
        self.index(0, 'src')
        self.key('src')
        df = conf.Conf()
        df.index(0, 'default')
        df.index(1, 'default_value')
        self.index(1,deal= df, dict_out=1)
        self.key('default', deal=df, dict_out=1)
        des = conf.Conf()
        des.index(0,'need')
        des.index(1, 'des')
        des.index(2, 'remove')
        self.index(2, deal=des, dict_out=1)

pass
class ArgSrcConf(conf.Conf):
    def init(self):
        super().init()
        self.index(0, 'key', need=1)
        self.key('key', 'val,value,data'.split(','), need=1)
        self.index(1, 'type')
        self.key('type', 'vtype'.split(","))
    def call(self, data, unit=None):
        data, upd = super().call(data, unit)
        val = data['key']
        if 'type' not in data:
            if type(val)==int:
                vtype = ArgType.list
            else:
                vtype = ArgType.dict
        else:
            vtype = data['type']
        data['type'] = ArgType.stand(vtype)
        return data, upd

class TrsArgsBuilder(Base):
    def init(self):
        self.conf = ArgItemConf()
        self.src_conf = ArgSrcConf()
    def visits(self, datas, vtype,obj):
        for key, item in dz.dict2iter(datas):
            item,_ = self.conf(item)
            it = ArgItem(key, vtype, *dz.g(item, need=0, default=0, default_value=None,des=None, remove=True))
            srcs = dz.g(item, src=[])
            if not dz.islist(srcs):
                srcs = [srcs]
            for src in srcs:
                src, _ = self.src_conf(src)
                key, vtype = src['key'], src['type']
                it.add(key, vtype)
            obj.add(it)
    def call(self, conf):
        lists = xf.g(conf, list={})
        dicts = xf.g(conf, dict={})
        if len(lists)+len(dicts)==0:
            return None
        keep,list_fill = xf.g(conf, keep=0, fill='exp')
        obj = TrsArgs([], keep, list_fill)
        self.visits(lists, ArgType.list, obj)
        self.visits(dicts, ArgType.dict, obj)
        return obj

class FullArgsBuilder(Base):
    def init(self):
        self.trs_build = TrsArgsBuilder()
        cf = conf.Conf()
        rcf = conf.Conf()
        rcf.index(0, 'base')
        rcf.index(1, 'last')
        rcf.index(2, 'min')
        cf.index(0, 'range', deal=rcf)
        cf.key('range', 'ranges'.split(","), deal=rcf)
        cf.index(1, 'list')
        cf.index(2, 'dict')
        cf.key('list', 'args,lists'.split(","))
        cf.key("dict", 'maps,dicts'.split(","))
        _cf = conf.Conf()
        _cf.index(0, 'keep')
        _cf.index(1, 'fill')
        _cf.key('fill', 'fill_type'.split(','))
        cf.index(3, deal=_cf, dict_out=1)
        self.conf = cf
    def call(self, conf):
        arr_args = ArrArgs()
        conf, upd = self.conf(conf)
        #print(f"BUILD conf: {conf}")
        ranges = dz.g(conf, range=None)
        if ranges is not None:
            base = ranges['base']
            if base is not None:
                last, min = dz.g(ranges, last=None,min=0)
                min = pyz.nnull(min, 0)
                range_args = RangeListArgs(base, last, min)
                arr_args.add(range_args)
        trs_args = self.trs_build(conf)
        if trs_args is not None:
            arr_args.add(trs_args)
        return arr_args



#coding=utf-8
import base64
import hashlib

try:
    import cryptography
except ModuleNotFoundError as exp:
    raise Exception("require package cryptography, install it by 'python3 -m pip install cryptography'")
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, padding
import os
class CryptBase:
    '''
        基础类，不可直接调用
    '''
    def __init__(self, pwd, coding = "utf-8"):
        self.coding = coding
        pwd = self.to_bytes(pwd)
        self.pwd = hashlib.md5(pwd).hexdigest().encode("ascii")
    def build(self, iv=None):
        self.iv = iv
        self.cipher = Cipher(algorithms.AES(self.pwd), modes.CBC(self.iv))
        self.padding = padding.PKCS7(algorithms.AES.block_size)
        self.hash_obj = hashlib.md5()
    def to_bytes(self, data):
        if type(data) == str:
            data = data.encode(self.coding)
        return data

class Encrypt(CryptBase):
    """
        md5 hash + aes 加密
        加密方式：
            aes(data+md5(data))
    """
    def __init__(self, pwd, iv=None, coding = "utf-8", iv2dt=True):
        super().__init__(pwd, coding)
        self.iv2dt = iv2dt
        if iv is None:
            iv = os.urandom(16)
        self.build(iv)
        self.cryptor = cryptor
        self.encryptor = self.cipher.encryptor()
        self.padder = self.padding.padder()
        self.hash_obj.update(cryptor.iv)
        self.first = 1
    def __call__(self, data=None, done=False):
        done = done or data is None
        if data is not None:
            data = self.update(data)
        else:
            data = b""
        if done:
            data += self.finalize()
        return data
    def update(self, data):
        data = self.cryptor.to_bytes(data)
        self.hash_obj.update(data)
        pdt = self.padder.update(data)
        out = self.encryptor.update(pdt)
        if self.first:
            self.first = 0
            if self.iv2dt:
                out = self.iv+out
        return out
    def finalize(self):
        hash = self.hash_obj.hexdigest().encode("ascii")
        pdt = self.padder.update(hash)+ self.padder.finalize()
        out = self.encryptor.update(pdt) + self.encryptor.finalize()
        return out

pass
class Decrypt(CryptBase):
    """
        md5 hash + aes 解密
        加密方式：
            aes解密成data+hash，校验md5(data)和hash是否一致
            aes(data+md5(data))
        update是解密数据，会保留解密后的最后32位不返回，留作最后当作hash值
    """
    def __init__(self, pwd, iv=None, coding = "utf-8"):
        super().__init__(pwd, coding)
        self.iv = iv
        if iv is not None:
            self.build(iv)
        self.remain = b""
    def __call__(self, data=None, done=False):
        done = done or data is None
        if data is not None:
            data = self.update(data)
        else:
            data = b""
        if done:
            data += self.finalize()
        return data
    def build(self, iv):
        super().build(iv)
        self.pwd = obj.hexdigest().encode("ascii")
        self.unpadder = self.padding.unpadder()
        self.decryptor = self.cipher.decryptor()
    def update(self, data):
        data = self.to_bytes(data)
        if self.iv is None:
            iv = data[:16]
            self.build(iv)
            data = data[16:]
            self.hash_obj.update(iv)
        pdt = self.decryptor.update(data)
        out = self.unpadder.update(pdt)
        out = self.remain+out
        if len(out)>32:
            self.remain = out[-32:]
            out = out[:-32]
            self.hash_obj.update(out)
        else:
            self.remain=out
            out = b""
        return out
    def finalize(self):
        pdt = self.decryptor.finalize()
        out = self.unpadder.update(pdt)+self.unpadder.finalize()
        out = self.remain+out
        self.remain = out[-32:]
        out = out[:-32]
        self.hash_obj.update(out)
        hash = self.hash_obj.hexdigest().encode("ascii")
        if self.remain != hash:
            raise Exception("Error pwd")
        return out

pass



class BufferInput:
    def __init__(self, buf):
        self.buf = buf
    def __call__(self, size=1):
        s = self.buf[:size]
        self.buf = self.buf[size:]
        return s

pass

class Buffer:
    def read_cache(self, size = 1):
        s = self.read(size)
        self.buffer+=s
        self.s_read = self.s_read[len(s):]
        return s
    def read(self, size = 1):
        if self.s_read is None:
            self.s_read = self.input(1)
        l = len(self.s_read)
        if l<size:
            self.s_read += self.input(size-l)
        rst = self.s_read[:size]
        if self.buffer is None:
            self.buffer = self.s_read[:0]
        return rst
    def clean2read(self, size = 1):
        self.s_read = self.s_read[size:]
        self.clean()
    def init(self):
        self.buffer = None
        self.s_read = None
    def __init__(self, input):
        self.input = input
        self.init()
    def add(self, size):
        if self.buffer is None:
            self.buffer = self.s_read[:size]
        else:
            self.buffer += self.s_read[:size]
    def size(self):
        if self.buffer is None:
            return 0
        return len(self.buffer)
    def full(self, size = 0, right = 1):
        return self.buffer
    def clean(self):
        self.buffer = self.buffer[:0]
    def rget(self, size=1):
        return self.buffer[-size:]
    def get(self, size=1):
        return self.buffer[:size]

pass

class StrBuffer:
    def read_cache(self, size = 1):
        s = self.read(size)
        l = len(s)
        self.buffer_size+=l
        self.read_base+=l
        return s
    def read(self, size = 1):
        s = self.str[self.read_base:self.read_base+size]
        return s
    def clean2read(self, size = 1):
        self.read_base+=size
        self.clean()
    def init(self):
        self.buffer_base = 0
        self.buffer_size = 0
        self.read_base = 0
        self.read_size = 0
        #self.buffer = None
        #self.s_read = None
    def __init__(self, s):
        self.str = s
        self.init()
    def add(self, size):
        # x = self.str[self.buffer_base+self.buffer_size:self.buffer_base+self.buffer_size+len(arr)]
        # if x!=arr:
        #     print(f"[ERROR] x:({x}), arr:({arr}), bb: {self.buffer_base}, bs:{self.buffer_size}, rb: {self.read_base}")
        #     print("[["+self.str[self.buffer_base:self.read_base+10]+"]]")
        #     raise Exception("")
        self.buffer_size+=size
    def size(self):
        return self.buffer_size
    def full(self):
        return self.str[self.buffer_base:self.buffer_base+self.buffer_size]
    def clean(self):
        self.buffer_base = self.read_base
        self.buffer_size = 0
    def rget(self, size=1):
        size = min(size, self.buffer_size)
        return self.str[self.buffer_base+self.buffer_size-size:self.buffer_base+self.buffer_size]
    def get(self, size=1):
        size = min(size, self.buffer_size)
        return self.str[self.buffer_base:self.buffer_base+size]

pass



#
from .base import Base
'''
    新的buildz.confz
'''
class Confz(Base):
    def init(self, fc_key = "fc", fn_key = "fn", objs="__confz.objs", fns = "__confz.fns", caches = "__confz.caches", fcs = "__confz.fcs", default_fn = "calls", obj_prefix = "", save_key = "save", default_save = True):
        '''
            初始化的入参除了default_save，都是存储数据的字段的key
        '''
        self.fc_key = fc_key
        self.fn_key = fn_key
        self.objs = objs
        self.fns = fns
        self.fcs = fcs
        self.caches = caches
        self.default_fn = default_fn
        self.obj_prefix=  obj_prefix
        self.default_save = default_save
        self.save_key = save_key
    def set(self, conf, id, obj, prefix=True):
        '''
            存放创建好的对象
        '''
        if prefix:
            id = self.obj_prefix+id
        conf(self.objs).set(id, obj)
    def get(self, conf, id):
        id = self.obj_prefix+id
        objs = conf(self.objs)
        obj = objs.get(id)
        if obj is None:
            obj = self.simple(conf, id)
        return obj
    def regist(self, curr, conf):
        '''
            fn: regist/reg
            name: fn_name
            path: fc_path
            call: false
        '''
        name, path, call = curr.gets("name, path, call")
        if not call:
            fns = conf(self.fns)
            fns.set(name, path)
            return
        fc = pyz.load(path)
        fc = fc(curr, conf)
        conf(self.fcs).set(name, fc)
    def calls(self, curr, conf):
        """
            fn: calls
            calls: [key1, key2, key3]
        """
        calls = curr.get("calls", [])
        rst = [self.simple(conf, k) for k in calls]
        return rst
    def init(self, conf):
        fcs = conf(self.fcs)
        fcs.set("regist", self.regist)
        fcs.set("reg", self.regist)
        fcs.set("calls", self.calls)
    def get_fc(self, curr, conf, default_fc):
        s_fc = curr.get(self.fc_key, None)
        if s_fc is not None:
            return pyz.load(s_fc)
        if default_fc is not None:
            return default_fc
        fn = curr.get(self,fn_key, self.default_fn)
        fcs = conf(self.fcs)
        fc = fcs.get(fn)
        if fc is not None:
            return fc
        fns = conf(self.fns)
        s_fc = fns.get(fn)
        if s_fc is None:
            return None
        fc = pyz.load(s_fc)
        fcs.set(fn, fc)
        return fc
    def simple(self, conf, key, default_fc = None):
        curr = conf(key)
        fc = self.get_fc(curr, conf, default_fc)
        obj = fc(curr, conf)
        if curr.get(self.save_key, self.default_save):
            self.set(key, obj,False)
        return obj
    def call(self, conf, key, default_fc=None):
        return self.simple(conf, key, default_fc)

pass
                
def wrap(fc, to_dict=0, unwrap=0, put_conf=1):
    '''
        封装函数，封装成Confz调用的方法的形式fc(curr, conf)
        fc(**dict): wrap(fc, to_dict=1, unwrap=1)
        fc(dict): wrap(fc, to_dict=1, unwrap=0, put_conf=0)
        fc(dict, conf): wrap(fc, to_dict=1, unwrap=0, put_conf=0)
        fc(curr): wrap(fc, to_dict=0, unwrap=0, put_conf=0)
        fc(curr, conf): wrap(fc, to_dict=0, unwrap=0, put_conf=1)
    '''
    if unwrap:
        assert to_dict==1
        assert put_conf==0
    def wfc(curr, conf):
        if to_dict:
            curr = curr.dict()
        if unwrap:
            return fc(**curr)
        elif not put_conf:
            return fc(curr)
        else:
            return fc(curr, conf)
    return wfc

pass



import socket, select, traceback, threading, time
from buildz import Base, logz, pyz

log = logz.simple()
class MiddleServer(Base):
    '''
        端口映射
        简单调用:
            python -m buildz.netz.tcp 监听ip 监听端口 目标ip 目标端口
    '''
    def init(self, addr, remote_addr, listen=5,wait=0.1, read_size=1024*1024):
        self.addr = tuple(addr)
        self.listen = listen
        self.remote_addr = tuple(remote_addr)
        self.read_size = read_size
        self.running = False
        self.wait = wait
        self.pairs = {}
        #self.lock = threading.Lock()
    def close_pair(self, a, b):
        #with self.lock:
        del self.pairs[id(a)]
        del self.pairs[id(b)]
        log("close").info(f"pair {a.getpeername()}, {b.getpeername()} close")
        for c in (a,b):
            try:
                c.close()
            except:
                pass
    def deal(self, skt):
        skt_read, skt_write = self.pairs[id(skt)]
        try:
            bts = skt_read.recv(self.read_size)
            if len(bts)==0:
                self.close_pair(skt_read, skt_write)
                return
            skt_write.send(bts)
            log("send").info(f"{skt_read.getpeername()} to {skt_write.getpeername()}: {len(bts)} bytes")
        except Exception as exp:
            log.debug(f"deal exp: {exp}")
            log.warn(f"traceback: {traceback.format_exc()}")
            self.close_pair(skt_read, skt_write)
    def connects(self):
        #with self.lock:
        arr = [v[0] for k,v in self.pairs.items()]
        return arr
    def add(self, skt_cli, addr):
        try:
            skt_srv = socket.socket()
            skt_srv.connect(tuple(self.remote_addr))
        except Exception as exp:
            log.error(f"add exp: {exp}")
            log.error(f"traceback: {traceback.format_exc()}")
            skt_cli.close()
            return
        log("connect").info(f"build pair from {skt_cli.getpeername()} to {skt_srv.getpeername()}")
        #with self.lock:
        self.pairs[id(skt_cli)] = (skt_cli, skt_srv)
        self.pairs[id(skt_srv)] = (skt_srv, skt_cli)
    def call(self, fc_stop = None):
        self.running=True
        if fc_stop is None:
            return self.waits_and_deals()
        else:
            threading.Thread(target=self.waits_and_deals,daemon=True).start()
            try:
                while not fc_stop():
                    time.sleep(1.0)
            finally:
                self.running=False
    def waits_and_deals(self):
        skt = socket.socket()
        skt.bind(self.addr)
        skt.listen(self.listen)
        self.skt = skt
        while self.running:
            skts = self.connects()
            skts.append(self.skt)
            (rlist,wlist,elist)=select.select(skts,[],[],self.wait)
            for skt in rlist:
                if skt == self.skt:
                    _skt,addr = self.skt.accept()
                    self.add(_skt, addr)
                else:
                    self.deal(skt)
    def waits(self):
        skt = socket.socket()
        skt.bind(self.addr)
        skt.listen(self.listen)
        self.skt = skt
        while self.running:
            skt,addr = self.skt.accept()
            self.add(skt, addr)
    def deals(self):
        while self.running:
            if len(self.pairs)==0:
                time.sleep(self.wait)
                continue
            (rlist,wlist,elist)=select.select(self.connects(),[],[],0)
            for skt in rlist:
                self.deal(skt)

pass

def test():
    import sys
    args = sys.argv[1:]
    ip = args.pop(0)
    port = int(args.pop(0))
    addr = (ip, port)
    ip = args.pop(0)
    port = int(args.pop(0))
    remote_addr = (ip, port)
    MiddleServer(addr, remote_addr)()

pass



#

from .lst import FcFpsListener
from .. import Base, xf, dz, pathz, pyz
import os, sys
from multiprocessing import Process
import subprocess
class Runner(Base):
    """
        监听目标文件的修改来运行目标代码，本代码通过子进程来运行目标代码，每次目标文件修改后，都会新建子进程运行目标代码，通过子进程运行目标代码主要考虑两点：
            1，目标文件可能就是包含目标代码的文件，或者是目标代码import的文件，当前进程已经import的模块是不会随着模块文件的修改而改变的，新建子进程来运行则不会有该问题
            2，python的线程无法被kill，在更新目标文件后，运行目标代码的线程如果没结束，要等其结束才能重新运行，子进程则可以被kill，监听到目标文件修改后，可以直接把子进程kill后重新创建子进程来运行目标代码
    """
    argv = []
    def init(self, fp, lst):
        fp = os.path.abspath(fp)
        self.dp = os.path.dirname(fp)
        self.path = pathz.Path(fp=self.dp)
        self.fp = fp
        self.lst = lst
        self.lst.set_update(self.update)
        self.lst.set_deal_exp(self.deal_exp)
        self.process = None
        self.exist_psutil = True
        self.reset()
    def loadf(self, fp):
        return xf.loadf(fp)
    def reset(self,last_update=False):
        conf = self.loadf(self.fp)
        fps, target, dp = dz.g(conf, fps = [], run=None, dp="")
        dp = self.path.fp(dp)
        sys.path.append(dp)
        self.path.set("sc", dp)
        fps = [self.path.sc(fp) for fp in fps]
        self.lst.clean()
        self.lst.add(self.fp, last_update)
        [self.lst.add(fp, last_update) for fp in fps]
        self.target=target
        self.conf = conf
    @staticmethod
    def process_argv():
        return Runner.argv
    @staticmethod
    def process_update(conf):
        target, sc_path, args, stdin, stdout, stderr = dz.g(conf, run=None,sc_path=__file__, args=[], stdin = "std", stdout = "std", stderr="std")
        encoding = dz.g(conf, encoding="utf-8")
        import sys, os
        args = [sc_path]+args
        Runner.argv = sys.argv
        sys.argv = args
        stds = [stdin, stdout, stderr]
        stds = [list(k) if type(k) in (list, tuple) else [k] for k in stds]
        modes = "rww"
        for std, mode in zip(stds, modes):
            if len(std)<3:
                std.append(mode)
        objs = []
        files = {}
        for std in stds:
            do_close = False
            if std[0]=='std':
                obj = sys.stdin if std[-1]=='r' else sys.stdout
            elif std[0]=='str':
                obj = io.StringIO(std[1])
            else:
                if std[1] not in files:
                    files[std[1]] = open(std[1], std[-1], encoding=encoding)
                obj = files[std[1]]
            objs.append(obj)
        sys.stdin = objs[0]
        sys.stdout = objs[1]
        sys.stderr = objs[2]
        try:
            fc = pyz.load(target)
            if callable(fc):
                fc()
        except Exception as exp:
            raise exp
        finally:
            for fp, f in files.items():
                f.flush()
    def kill_chs(self, pid):
        if not self.exist_psutil:
            return
        try:
            import psutil
        except ModuleNotFoundError as exp:
            self.exist_psutil = False
            print(f"[WARN] psutil not installed, children process of subprocess will not being killed")
            print(f"[WARN] if children processes exists and need to be killed, do 'pip install psutil' by yourself and restart")
            print(f"[警告] 没有装psutil，本代码无法删除目标进程的子进程，如果需要本代码删除子进程，需要装psutil然后重启: pip install psutil")
            return
        try:
            parent_process = psutil.Process(pid)
            children = parent_process.children(recursive=True)
            for child in children:
                child.kill()
        except psutil.NoSuchProcess as exp:
            pass
            #print(f"[WARN] Process {pid} kill error: {exp}, maybe process already end")
    def process_command(self):
        cmd = f"python -m buildz.sc.subchild {self.fp}"
        return cmd.split(" ")
    def update(self, fps):
        if self.process is not None:
            self.kill_chs(self.process.pid)
            self.process.kill()
            self.process = None
        self.reset(True)
        p = subprocess.Popen(self.process_command())
        #p = Process(target = self.process_update,args=[self.conf], daemon=True)
        #p.start()
        self.process = p
    def deal_exp(self, exp, fmt_exc):
        print(f"exp: {exp}")
        print(f"traceback: \n{fmt_exc}")

def test():
    fp = sys.argv[1]
    lst = FcFpsListener()
    runner = Runner(fp, lst)
    lst.run()

pass
pyz.lc(locals(), test)


from .. import base
from .. import item
from .. import exp
from . import lr
from ...base import Args
class ListMapDeal(lr.LRDeal):
    """
        分隔符，有分隔符后将缓存的数据当作字符串
    """
    def init(self, left, right, as_map = False, as_args=False):
        super().init(left, right, "listmap")
        self.as_map = as_map
        self.as_args = as_args
    def types(self):
        return ['list']
    def build(self, obj):
        # if self.check_right(obj):
        #     return None
        val = obj.val
        if len(val)==0:
            obj.val = []
            obj.is_val = 1
            return obj
        return self.build_arr(val, obj.pos)
    def build_arr(self, arr, arr_pos):
        rst = {}
        lst = []
        mp = {}
        i = 0
        while i<len(arr):
            obj = arr[i]
            opt = None
            if i+1<len(arr):
                opt = arr[i+1]
            if opt is not None and opt.type == 'kv':
                if i+2>=len(arr):
                    raise exp.Exp(f"u f in listmap: {arr}", arr_pos)
                val = arr[i+2]
                if type(obj.val)==list:
                    obj.val = tuple(obj.val)
                mp[obj.val]=val.val
                i+=3
            else:
                lst.append(obj.val)
                i+=1
        if self.as_args:
            return item.Item(Args(lst, mp), arr_pos, type='args', is_val=1)
        if len(mp)==0:
            if self.as_map:
                if len(lst)==0:
                    return item.Item(mp, arr_pos, type="map", is_val=1)
                else:
                    return item.Item(Args(lst, mp), arr_pos, type='args', is_val=1)
            return item.Item(lst, arr_pos, type="list", is_val=1)
        elif len(lst)==0:
            return item.Item(mp, arr_pos, type="map", is_val=1)
        else:
            return item.Item(Args(lst, mp), arr_pos, type='args', is_val=1)

pass


#coding=utf-8
from ..ioc.base import Base, EncapeData
from .base import FormatData,FormatDeal
from buildz import xf, pyz
import os
dp = os.path.dirname(__file__)
join = os.path.join
class JoinDeal(FormatDeal):
    """
    文件路径合并join:
        {
            id:id
            type: join
            data: [...]
        }
    简写:
        [[join, id], data]
        [join, data]
    例:
        [join, [[val, home], [val, buildz]]] //返回字符串 "home/buildz"
    """
    def init(self, fp_lists=None, fp_defaults=None):
        super().init("JoinDeal", fp_lists, fp_defaults, join(dp, "conf", "join_lists.js"), None)
    def deal(self, edata:EncapeData):
        data = edata.data
        data = self.fill(data)
        lists = xf.g1(data, join = [], data=[])
        conf = edata.conf
        rst = [self.get_obj(k, conf, edata.src, edata.info) for k in lists]
        return join(*rst)

pass





from buildz.tools import *
from buildz import tz, argx, dz
import os
join = os.path.join
import struct
r'''
python -m buildz.tz.filediff diff src_dp target_dp diff_file
比较src_dp和target_dp的差异，把target_dp新增的内容记录到diff_file中

python -m buildz.tz.filediff recover src_dp target_dp diff_file
根据diff_file和src_dp，将新增的文件和修改过内容的文件添加到target_dp中（没改过的不会添加）
'''
def diff(fp1, fp2):
    bs1 = fz.read(fp1).decode("utf-8")
    bs2 = fz.read(fp2).decode("utf-8")
    stps = tz.m_steps(bs1, bs2,split=0)
    return tz.m_encode(stps)
def diffs(dp1, dp2):
    fps1 = set(fz.search(dp1, relative=1))
    fps2 = set(fz.search(dp2, relative=1))
    fps = set(fps1)
    fps.update(fps2)
    rst = []
    for fp in fps:
        if fp in fps1 and fp in fps2:
            fp1 = join(dp1, fp)
            fp2 = join(dp2, fp)
            h1 = fz.fhash(fp1)
            h2 = fz.fhash(fp2)
            if h1!=h2:
                rst.append(['ne', fp, diff(fp1, fp2)])
            pass
        elif fp in fps1:
            rst.append(['del', fp])
        else:
            rst.append(['add', fp, fz.read(join(dp2, fp))])
    return rst
tps = 'ne,del,add'.split(",")
btps = b'n,d,a'.split(b",")
itps = {}
for i in range(3):
    itps[tps[i]] = i
    itps[btps[i]] = i
def rst2bts(rst):
    n = len(rst)
    bs = b''
    bn = struct.pack(">H", n)
    bs = bn
    for it in rst:
        tp,fp=it[:2]
        btp = btps[itps[tp]]
        bfp = bytes([len(fp)])+fp.encode("utf-8")
        tmp = btp+bfp
        if len(it)>2:
            _bs = it[2]
            bbs = struct.pack(">I", len(_bs))+_bs
            tmp+=bbs
        bs+=tmp
    return bs
def bts2rst(bs):
    bn = bs[:2]
    bs = bs[2:]
    n = struct.unpack(">H", bn)[0]
    rst = []
    for i in range(n):
        btp = bs[:1]
        tp = tps[itps[btp]]
        lfp = bs[1:2][0]
        fp = bs[2:2+lfp].decode("utf-8")
        bs = bs[2+lfp:]
        it = [tp, fp]
        if tp in 'ne,add'.split(","):
            l_bs = struct.unpack(">I", bs[:4])[0]
            _bs = bs[4:4+l_bs]
            bs = bs[4+l_bs:]
            it.append(_bs)
        rst.append(it)
    return rst

def fdiffs(dp1, dp2, fp):
    rst = diffs(dp1, dp2)
    bs = rst2bts(rst)
    fz.write(bs, fp)
def fupds(dp1, fp, dp2):
    bs = fz.read(fp)
    rst = bts2rst(bs)
    upds(dp1, rst, dp2)
def upd(fp1, bs_step, fp2):
    bs1 = fz.read(fp1).decode("utf-8")
    stps = tz.m_decode(bs_step)
    bs2 = tz.m_update(bs1, stps,split=0).encode("utf-8")
    fz.write(bs2, fp2)
def frp(fp):
    return fp.replace("\\","/")
def upds(dp1, rst, dp2):
    for item in rst:
        tp,fp=item[:2]
        if len(item)>2:
            bs = item[2]
        if tp == 'del':
            fz.removes(join(dp2, fp))
        elif tp == 'add':
            fz.makefdir(join(dp2, fp))
            fz.write(bs, join(dp2, fp))
        elif tp=='ne':
            fz.makefdir(join(dp2, fp))
            upd(join(dp1, fp), bs, join(dp2, fp))

pass
fetch = argx.Fetch(*xf.loads("[order, src, tgt, fp],{o:order,s:src,t:tgt,f:fp}"))

def test():
    conf = fetch()
    order,src,tgt,fp = dz.g(conf, order=0,src=0,tgt=0,fp=0)
    if order == 'diff':
        print("do diff")
        fdiffs(src, tgt, fp)
    elif order == 'recover':
        print("do recover")
        fupds(src, fp, tgt)
    print("done")

pyz.lc(locals(), test)


import socket
class Broad:
    '''
        局域网广播
    '''
    def __init__(self, port: int=5005):
        self.port = port
        # 创建UDP套接字
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # 启用广播模式
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.skt = sock
        self.bind = False
    def recv(self, size=1024*1024):
        if not self.bind:
            self.skt.bind(('', self.port))
            self.bind=True
        data, addr = self.skt.recvfrom(size)
        return data, addr
    def send(self, message: bytes, port=None):
        broadcast_addr = '<broadcast>'
        port = port or self.port
        self.skt.sendto(message, (broadcast_addr, port))
    def close(self):
        self.skt.close()

pass
def test():
    import sys, time
    sec=2.0
    od = sys.argv[1].strip().lower()[:1]
    if len(sys.argv)>2:
        sec = float(sys.argv[2])
    broad=Broad()
    try:
        if od == 's':
            while True:
                broad.send(b"test from broadcast")
                print(f"send success")
                time.sleep(sec)
        elif od == 'r':
            while True:
                msg, addr = broad.recv()
                print(f"recive from {addr}: {msg.decode('utf-8')}")
        else:
            assert 0
    finally:
        broad.close()
if __name__=="__main__":
    test()



#

from .jsc import fetch_methods_text, rps_dct, fetch_vals
def make_fcs(fc, add_src=True):
    def wfcs(codes, args, text):
        fcs, has_def = fetch_methods_text(codes, text)
        rst = []
        others = []
        if has_def:
            if add_src:
                rst.append(codes.rstrip())
            else:
                others.append(codes.rstrip())
        for _type, method, params, exp in fcs:
            a, b = fc(_type, method, params, exp, args)
            rst+=a
            others+=b
        return rst, others
    return wfcs


import socket,threading
from . import mhttp
from . import record as _record
from buildz import Base
import traceback
log = mhttp.log
class ProxyDealer(Base):
    """
        http和https代理
    """
    def init(self, skt, channel_read_size=1024000, record=None, protocol = "HTTP/1.1", default_deal=None):
        self.wskt = mhttp.WSocket.Bind(skt)
        self.channel_read_size = channel_read_size
        self.deals = {}
        if default_deal is None:
            default_deal = self.default_deal
        self.deals[None] = default_deal
        self.deals['CONNECT'] = self.connect_deal
        self.skts = {}
        self.protocol = protocol
        if record is None:
            record = _record.MsgLog(log)
        self.record = record
    def simple_chunked(self, skt_read, skt_send=None):
        while True:
            n, dt = mhttp.chunked_data(skt_read)
            self.record.add_chunk(dt)
            bts = mhttp.chunked_encode(dt)
            if skt_send is not None:
                skt_send.send(bts)
            if n==0:
                break
    def do_connect(self, addr):
        return mhttp.WSocket.Connect(addr)
    def default_deal(self, skt_cli, line, headers, data_size, skt=None, fc_connect=None):
        if fc_connect is None:
            fc_connect = self.do_connect
        self.record.request(line, headers, data_size)
        http_type, url, protocol = line
        skt,fc_done = mhttp.http_send_head(http_type, url, headers, data_size,protocol, skt, self.skts, fc_connect)
        if data_size>0:
            data = skt_cli.read(data_size)
            skt.send(data)
            self.record.add(data)
        if mhttp.check_chunked(headers):
            self.simple_chunked(skt_cli, skt)
        self.record.finish()
        line, headers, data_size=mhttp.http_recv(skt)
        self.record.response(line, headers, data_size)
        protocol, code, rsp_text = line
        bts = mhttp.http_encode_rsp(code, rsp_text, headers, data_size, protocol)
        skt_cli.send(bts)
        if data_size>0:
            data = skt.read(data_size)
            skt_cli.send(data)
            self.record.add(data)
        if mhttp.check_chunked(headers):
            self.simple_chunked(skt, skt_cli)
        self.record.finish()
        fc_done()
    def close(self):
        self.wskt.close()
        for addr,skt in self.skts.items():
            try:
                skt.close()
            except:
                pass
    def error(self, skt_cli, code=404, txt="Not Found", data=None):
        if data is None:
            data = txt
        if type(data)==str:
            data = data.encode("utf-8")
        bts = mhttp.http_encode_rsp(code, txt, data_size=len(data), protocol = self.protocol)
        skt_cli.send(bts+data)
        log.debug(f"error: {code}, txt={txt}")
    def connect_deal(self, skt_cli, line, headers, data_size, skt=None):
        self.record.request(line, headers, data_size)
        if data_size>0:
            dt = skt_cli.read(data_size)
            self.record.add(dt)
        self.record.finish()
        http_type, url, protocol = line
        addr = url.split(":")
        if len(addr)==1:
            addr.append(80)
        addr[1] = int(addr[1])
        addr = tuple(addr)
        code, txt = 200, "OK"
        need_close = skt is None
        if skt is None:
            try:
                skt = mhttp.WSocket.Connect(addr)
            except:
                code,txt=404,"Not Found"
        self.record.response([self.protocol,code, txt],{},0).finish()
        if code!=200:
            return self.error(skt_cli, code, txt)
        bts = mhttp.http_encode_rsp(code, txt, protocol = self.protocol)
        skt_cli.send(bts)
        try:
            self.deal_channel(skt_cli, skt)
        finally:
            if need_close:
                skt.close()
    def deal_channel(self, skt_cli, skt_srv):
        return self.direct_channel(skt_cli, skt_srv)
    def direct_channel(self, skt_cli, skt_srv):
        try:
            while True:
                while skt_cli.readable():
                    bts = skt_cli.recv(self.channel_read_size)
                    if len(bts)==0:
                        return
                    skt_srv.send(bts)
                while skt_srv.readable():
                    bts = skt_srv.recv(self.channel_read_size)
                    if len(bts)==0:
                        return
                    skt_cli.send(bts)
        except Exception as exp:
            log.debug(f"channel exp: {exp}")
            log.warn(f"traceback: {traceback.format_exc()}")
    def deal(self):
        try:
            if not self.wskt.readable():
                return True
            line, headers, data_size = mhttp.http_recv(self.wskt.rfile)
            if line is None:
                return True
            log.debug(f"[DEAL] START")
            log.debug(f"proxy recv: {line, headers, data_size}")
        except Exception as exp:
            log.warn(f"http_recv exp: {exp}")
            log.warn(f"traceback: {traceback.format_exc()}")
            return False
        http_type, url, protocol = line
        if http_type not in self.deals:
            http_type = None
        fc = self.deals[http_type]
        fc(self.wskt, line, headers, data_size)
        log.debug(f"[DEAL] END")
        return True
    def call(self):
        log.debug(f"[TESTZ] new deal")
        try:
            while True:
                if not self.deal():
                    break
        finally:
            self.wskt.close()
            #self.monitor.close()

pass

class Proxy(Base):
    def init(self, addr, listen=5, record=None):
        self.addr = addr
        self.listen = listen
        self.ths = []
        self.running=False
        if record is None:
            record = _record.MsgLog(log)
        self.record = record
    def close(self):
        self.skt.close()
    def make_dealer(self, skt, addr):
        return ProxyDealer(skt, record=self.record.clone())
    def call(self, wait_time=0.1):
        self.running=True
        skt = socket.socket()
        skt.bind(self.addr)
        skt.listen(self.listen)
        self.skt = skt
        try:
            while self.running:
                while not mhttp.readable(self.skt,wait_time):
                    pass
                skt,addr = self.skt.accept()
                deal = self.make_dealer(skt, addr)
                th = threading.Thread(target=deal,daemon=True)
                th.start()
                self.ths.append(th)
        finally:
            self.close()



#

from .conf import mg
from .. import xf
from .ioc.base import Params
from .wrap import wraps, default_wraps
s_default_conf = r"""
confs.pri: {
    deal_obj:{
        type=deal
        src=<buildz>.iocz.conf_deal.obj.ObjectDeal
        deals: [obj,object]
        call=1
    }
    deal_val:{
        type=deal
        src=<buildz>.iocz.conf_deal.val.ValDeal
        deals: [val,value]
        call=1
    }
    deal_ref: {
        type:deal
        src:<buildz>.iocz.conf_deal.ref.RefDeal
        deals: ref
        call=1
    }
    deal_ioc: {
        type:deal
        src:<buildz>.iocz.conf_deal.ioc.IOCDeal
        deals: ioc
        call=1
    }
    deal_cvar: {
        type=deal
        src: <buildz>.iocz.conf_deal.cvar.CVarDeal
        deals: cvar
        call=1
    }
    deal_call: {
        type=deal
        src: <buildz>.iocz.conf_deal.call.CallDeal
        deals: (call, fc)
        call=1
    }
    deal_env: {
        type=deal
        src: <buildz>.iocz.conf_deal.env.EnvDeal
        deals: (env, profile,conf)
        call=1
    }
    eref: {
        type=obj
        src: <buildz>.iocz.conf_deal.ref.RefDeal
        args: (0)
    }
    deal_eref: {
        type=deal
        src: [ref, eref]
        deals: (eref)
        call=0
    }
    refe: {
        type=obj
        src: <buildz>.iocz.conf_deal.ref.RefDeal
        args: (1)
    }
    deal_refe: {
        type=deal
        src: [ref, refe]
        deals: (refe)
        call=0
    }
    deal_method: {
        type=deal
        src=<buildz>.iocz.conf_deal.method.MethodDeal
        deals=(method,obj_fc,o_fc)
        call=1
    }
    deal_attr: {
        type=deal
        src=<buildz>.iocz.conf_deal.attr.AttrDeal
        deals=(attr, obj_var, o_var)
        call=1
    }
}
builds: [deal_obj,deal_val,deal_ref,deal_ioc,deal_cvar, deal_env,deal_call, deal_eref, deal_refe,deal_method,deal_attr]
""".replace("<buildz>", "buildz")
def build(conf = None, default_conf = True, default_wrap = True):
    global s_default_conf
    if default_conf and type(default_conf) not in (str, list, tuple, dict):
        default_conf = s_default_conf
    if type(default_conf) == str:
        default_conf = xf.loads(default_conf)
    obj = mg.ConfManager(conf)
    if default_conf:
        obj.add_conf(default_conf)
    if default_wrap:
        get_wraps().bind(obj)
    return obj

pass

def build_wraps(default=True):
    obj = wraps.WrapUnits()
    if default:
        default_wraps.build(obj)
    return obj

global_wraps = None
def get_wraps():
    global global_wraps
    if global_wraps is None:
        global_wraps = build_wraps()
    return global_wraps

def get_wrap(ns=None):
    return get_wraps()(ns)


#

#from buildz.argz import conf_argz as argz
from buildz.argz import conf_callz as callz
from buildz.tls import *
conf = r"""
(
    (argz, main)
    (search, test_ret,test_var,test_env)
)
{
    type: argz
    id: search
    call: search_fc
    judge: eq(args[0], search)
    conf={
        range: 1
        dict: {
            dp: ((0, dp), false, (null,1))
            pt_fp: ((1, pt_fp), (true,null))
            pt: ((2, pt), (true,null))
        }
    }
}
[
    (argz_ret, test_ret)
    [1]
    eq(args[0], test)
]
[
    (argz_env, test_env)
    [1]
    eq(args[0], env)
]
[
    (argz, test_var)
    test_fc
    eq(args(0), var)
]
[
    (cvar, search_fc)
    buildz.fz.search
]
[
    (obj, search_obj)
    buildz.fz.search
    [(ref, dp)]
]
(
    (vargz, vmain)
    (vsearch)
)
(
    (val, test_var),123
)
{
    type: vargz
    id: vsearch
    call: search_obj
    judge: and(eq(args[0], search),eq(ref(test_var),123))
    conf={
        range: 1
        dict: {
            dp: ((0, dp), false, (null,1))
            pt_fp: ((1, pt_fp), (true,null))
            pt: ((2, pt), (true,null))
        }
    }
}
"""
def test_fc(*a, **b):
    print(f"test_fc({a}, {b})")
    return a,b
def test():
    args = ['search','.']
    maps = {
        'pt_fp': r".*\.py$"
    }
    bd = argz.build()
    bd.adds(conf)
    bd.push_var("test_fc",test_fc)
    #rst = bd("main", args, maps)
    rst = bd.get("vmain", args, maps)
    print(f"rst: {rst}")
    print(bd.mg.envs)
    pass

pass

pyz.lc(locals(),test)


#
from .tdict import TagDict
from ... import pyz, dz,Base
class Key:
    Pub = "pub"
    Pri = "pri"
    Ns = "ns"
    tags_ns = [Ns,Pub]
    tags_pub = [Pub]
    tags_pri = None
    @staticmethod
    def is_pub(s):
        return s in ("pub", "public")
    @staticmethod
    def is_pri(s):
        return s in ("pri", "prv", "private")
    @staticmethod
    def is_ns(s):
        return s in ("ns", "namespace")
    @staticmethod
    def stand(s):
        if Key.is_pub(s):
            return Key.Pub
        elif Key.is_ns(s):
            return Key.Ns
        elif Key.is_pri(s):
            return Key.Pri
        return None
pass
class UnitBase(Base):
    def init(self, ns=None, id=None):
        self.ns = ns
        self.id = id
class TagData(TagDict):
    '''
        分成三个域，公共pub，私有pri和同域名ns
        设置访问规则
        不同域名只能访问pub数据
        同域名访问pub和ns
        同一个配置文件内（同一个id）访问pub，ns和pri
    '''
    Key = Key
    def init(self, ns=None, id=None):
        self.ns = ns
        self.id = id
        super().init(Key.Pub)
    def set_id(self, id):
        self.id = id
    def call(self, *a,**b):
        return self.tget(*a,**b)
    @staticmethod
    def nsid(src, id):
        if isinstance(src, TagData) or isinstance(src, UnitBase):
            ns = src.ns
            id = src.id
        else:
            ns = src
        return ns, id
    def tget(self, key, src=None, id=None):
        ns, id = self.nsid(src, id)
        if id == self.id:
            tags = Key.tags_pri
        elif ns == self.ns:
            tags = Key.tags_ns
        else:
            tags = Key.tags_pub
        return self.get(key, tags)

pass




from .base import *
from ..ioc_deal import ioc
class IOCDeal(ioc.IOCDeal):
    def init(self):
        super().init()
        conf = BaseConf()
        conf.index(1, 'target')
        conf.key('target', 'data,key'.split(","))
        self.update = conf

pass


#

from .tdata import TagData
from .base import *
from ... import dz
class Dataset(Base):
    '''
        所有数据集，只包含pub和ns部分
    '''
    def str(self):
        return f"Dataset(pub={self.pub}, ns={self.ns})"
    def init(self, ids):
        self.ids = ids
        self.pub = {}
        self.ns = {}
        self.objs = {}
    def update(self, maps, ns=None, tag=None, id=None, flush=False):
        if flush:
            maps = dz.unflush_maps(maps, self.ids.id)
        for key, val in maps.items():
            self.set(key, val, ns, tag, id)
    def set(self, key, val, ns=None, tag=None, id=None):
        if tag == TagData.Key.Ns:
            self.ns_set(key,val,ns,id)
        else:
            self.pub_set(key,val,ns,id)
    def pub_set(self, key, val, ns=None, id=None):
        keys = self.ids(ns)+self.ids(key)
        dz.dset(self.pub, keys, val)
    def ns_set(self, key, val, ns=None, id=None):
        map = dz.get_set(self.ns, ns, dict())
        dz.dset(map, self.ids(key),val)
        #map[key] = val
    def add(self, data):
        ns = data.ns
        self.objs[data.id] = data
        for key, val in data.tag(data.Key.Pub).items():
            self.pub_set(key, val, ns, data)
        for key, val in data.tag(data.Key.Ns).items():
            self.ns_set(key, val, ns, data)
        if data.dts!=self:
            data.bind(self)
    def tget(self, key, ns=None, tag=None, id=None):
        obj, keys,tag, find = self.get(key,ns,tag, id)
        return obj, tag, find
    def get(self, key, ns=None, tag=None, id = None):
        keys = self.ids(key)
        if len(keys)==0:
            return None,None,None,0
        if tag == TagData.Key.Pub:
            find =False
        else:
            obj, find = self.ns_get(key,ns,id)
            tag = TagData.Key.Ns
            keys = key
        if not find:
            obj, keys, find=self.pub_get(key,ns,id)
            tag=TagData.Key.Pub
        return obj,keys,tag,find
    def _pub_get(self, key, ns=None, id=None):
        ids_key = self.ids(key)
        ids = ids_key
        obj,find = dz.dget(self.pub, ids)
        if find:
            return obj, ids, find
        if ns is None:
            return 0,0, 0
        ids = self.ids(ns) + ids_key
        obj,find = dz.dget(self.pub, ids)
        if find:
            return obj, ids, find
        return None,0,0
    def pub_get(self, key, ns=None, id=None):
        obj,keys,find = self._pub_get(key, ns, id)
        if find:
            return obj,keys,find
        return None,None,0
    def ns_get(self, key, ns=None, id=None):
        map = dz.get_set(self.ns, ns, dict())
        obj, find = dz.dget(map, self.ids(key))
        if not find:
            return None,0
        return obj,1
